from typing import Mapping
import inspect
import math
import numbers

from visidata import vd, asyncthread, ENTER, deduceType, anytype
from visidata import Sheet, Column, VisiData, ColumnItem, TableSheet, BaseSheet, Progress, ColumnAttr, SuspendCurses, TextSheet, setitem
import visidata

vd.option('visibility', 0, 'visibility level')
vd.option('default_sample_size', 100, 'number of rows to sample for regex.split (0=all)', replay=True)
vd.option('fmt_expand_dict', '%s.%s', 'format str to use for names of columns expanded from dict (colname, key)') #1457
vd.option('fmt_expand_list', '%s[%s]', 'format str to use for names of columns expanded from list (colname, index)')


class PythonSheet(Sheet):
    def openRow(self, row):
        return PyobjSheet("%s[%s]" % (self.name, self.rowname(row)), source=row)

class PythonAtomSheet(PythonSheet):
    '''a sheet to display one Python object that does not offer deeper inspection,
        like None, a bool, or an int/float'''
    rowtype = 'object'  #singular, because it should only ever hold one
    columns = [
        Column('value', getter=lambda col,row: row,
                        setter=lambda c,r,v: None)
    ]
    def loader(self):
        self.rows = [self.source]
        self.column('value').type = deduceType(self.source)

    def openRow(self, row):
        vd.fail('cannot dive deeper on this object')
    def openCell(self, col, row, rowidx=None):
        vd.fail('cannot dive deeper on this object')
    def openRowPyobj(self, rowidx):
        vd.fail('cannot dive deeper on this object')
    def openCellPyobj(self, col, rowidx):
        vd.fail('cannot dive deeper on this object')
    def newRow(self):
        vd.fail('adding rows to this sheet is not supported')

#### generic list/dict/object browsing
@VisiData.global_api
def view(vd, obj):
    vd.run(PyobjSheet(getattr(obj, '__name__', ''), source=obj))


def PyobjColumns(obj):
    'Return columns for each public attribute on an object.'
    cols = []
    for k in dir(obj):
        coltype = anytype
        try:
            if k.startswith('_') or callable(getattr(obj, k)):
                continue
            coltype = deduceType(getattr(obj, k))
        except AttributeError: #2631 attributes like formatted_help can raise AttributeError
            pass
        cols.append(ColumnAttr(k, type=coltype))
    return cols

def AttrColumns(attrnames):
    'Return column names for all elements of list `attrnames`.'
    return [ColumnAttr(name) for name in attrnames]


def SheetList(*names, **kwargs):
    'Creates a Sheet from a list of homogeneous dicts or namedtuples.'

    src = kwargs.get('source', None)
    if not src:
        vd.warning('no content in %s' % names)
        return Sheet(*names, **kwargs)

    if isinstance(src[0], Mapping):
        return ListOfDictSheet(*names, **kwargs)
    elif isinstance(src[0], tuple):
        if getattr(src[0], '_fields', None):  # looks like a namedtuple
            return ListOfNamedTupleSheet(*names, **kwargs)

    # simple list
    return ListOfPyobjSheet(*names, **kwargs)

class ListOfPyobjSheet(PythonSheet):
    rowtype = 'python objects'
    def loader(self):
        self.rows = self.source
        self.columns = []
        self.addColumn(Column(self.name,
                               getter=lambda col,row: row,
                               setter=lambda col,row,val: setitem(col.sheet.source, col.sheet.source.index(row), val)))

        for c in PyobjColumns(self.rows[0]):
            self.addColumn(c)

        if len(self.columns) > 1:
            self.columns[0].width = 0

# rowdef: dict
class ListOfDictSheet(PythonSheet):
    rowtype = 'dicts'
    def reload(self):
        self.columns = []
        self._knownKeys = set()
        for row in self.source:
            for k in row:
                if k not in self._knownKeys:
                    self.addColumn(ColumnItem(k, k, type=deduceType(row[k])))
                    self._knownKeys.add(k)
        self.rows = self.source

# rowdef: namedtuple
class ListOfNamedTupleSheet(PythonSheet):
    rowtype = 'namedtuples'
    def reload(self):
        self.columns = []
        for i, k in enumerate(self.source[0]._fields):
            self.addColumn(ColumnItem(k, i))
        self.rows = self.source


# rowdef: PyObj
class SheetNamedTuple(PythonSheet):
    'a single namedtuple, with key and value columns'
    rowtype = 'values'
    columns = [ColumnItem('name', 0), ColumnItem('value', 1)]

    def __init__(self, *names, **kwargs):
        super().__init__(*names, **kwargs)

    def reload(self):
        self.rows = list(zip(self.source._fields, self.source))

    def openRow(self, row):
        return PyobjSheet(f'{self.name}.{row[0]}', source=row[1])


# source is dict
class SheetDict(PythonSheet):
    rowtype = 'items'  # rowdef: keys
    columns = [
        Column('key'),
        Column('value', getter=lambda c,r: c.sheet.source[r],
                        setter=lambda c,r,v: setitem(c.sheet.source, r, v)),
    ]
    nKeys = 1
    def reload(self):
        self.rows = list(self.source.keys())

    def openRow(self, row):
        return PyobjSheet(f'{self.name}.{row}', source=self.source[row])


class ColumnSourceAttr(Column):
    'Use row as attribute name on sheet source'
    def calcValue(self, attrname):
        return getattr(self.sheet.source, attrname)
    def setValue(self, attrname, value, setModified=True):
        ret = setattr(self.sheet.source, attrname, value)
        if setModified:
            self.sheet.setModified()
        return ret

def docstring(obj, attr):
    v = getattr(obj, attr)
    if callable(v):
        return v.__doc__
    return '<type %s>' % type(v).__name__

# rowdef: attrname
class PyobjSheet(PythonSheet):
    'Generic Sheet for any Python object.  Return specialized subclasses for lists of objects, namedtuples, and dicts.'
    rowtype = 'attributes'
    columns = [
        Column('attribute'),
        ColumnSourceAttr('value'),
        Column('signature', width=0, getter=lambda c,r: dict(inspect.signature(getattr(c.sheet.source, r)).parameters)),
        Column('docstring', getter=lambda c,r: docstring(c.sheet.source, r))
    ]
    nKeys = 1

    def __new__(cls, *names, **kwargs):
        'Return Sheet object of appropriate type for given sources in `args`.'
        pyobj=kwargs.get('source', object())
        if pyobj is None:
            return None
        elif isinstance(pyobj, numbers.Number):
            return PythonAtomSheet(*names, source=pyobj)
        elif isinstance(pyobj, (list, tuple)):
            if getattr(pyobj, '_fields', None):  # list of namedtuple
                return SheetNamedTuple(*names, **kwargs)
            else:
                return SheetList(*names, **kwargs)
        elif isinstance(pyobj, set):
            return ListOfPyobjSheet(*names, source=list(pyobj))
        elif isinstance(pyobj, Mapping):
            return SheetDict(*names, **kwargs)
        elif isinstance(pyobj, str):
            return TextSheet(*names, source=pyobj.splitlines())
        elif isinstance(pyobj, bytes):
            return TextSheet(*names, source=pyobj.decode(cls.options.encoding).splitlines())
        elif isinstance(pyobj, object):
            obj = super().__new__(cls)  #, *names, **kwargs)
            return obj
        else:
            vd.error("cannot load '%s' as pyobj" % type(pyobj).__name__)

    def reload(self):
        self.rows = []
        vislevel = self.options.visibility
        for r in dir(self.source):
            # reading these attributes can cause distracting fail() messages
            if r in ('onlySelectedRows', 'someSelectedRows'):
                vd.warning(f'skipping attribute: {r}')
                continue
            try:
                if vislevel <= 2 and r.startswith('__'): continue
                if vislevel <= 1 and r.startswith('_'): continue
                if vislevel <= 0 and callable(getattr(self.source, r)): continue
            except Exception:
                pass

            self.addRow(r)

    def openRow(self, row):
        'dive further into Python object'
        v = getattr(self.source, row)
        return PyobjSheet(self.name + "." + str(row), source=v() if callable(v) else v)


@TableSheet.api
def openRow(sheet, row, rowidx=None):
    'Return Sheet diving into *row*.'
    if row is None or sheet.nRows == 0: vd.fail('no row to dive into')
    if rowidx is None:
        k = sheet.rowname(row) or str(sheet.cursorRowIndex)
    else:
        k = rowidx

    name = f'{sheet.name}[{k}]'
    return TableSheet(name,
                      rows=sheet.visibleCols,
                      sourceRow=sheet.cursorRow,
                      columns=[
                        Column('column', getter=lambda c,r: r.name),
                        Column('value', getter=lambda c,r: r.getTypedValue(c.sheet.sourceRow), setter=lambda c,r,v: r.setValue(c.sheet.sourceRow, v)),
                      ],
                      nKeys=1)

@TableSheet.api
def openCell(sheet, col, row, rowidx=None):
    'Return Sheet diving into cell at *row* in *col*.'
    if col is None or row is None or sheet.nRows == 0: vd.fail('no cell to dive into')
    if rowidx is None:
        k = sheet.rowname(row) or str(sheet.cursorRowIndex)
    else:
        k = rowidx
    name = f'{sheet.name}[{k}].{col.name}'
    return PyobjSheet(name, source=col.getTypedValue(row))

@Sheet.api
@asyncthread
def openRows(sheet, rows):
    for r in Progress(rows):
        vd.push(sheet.openRow(r))

    vd.sync()

@Sheet.api
@asyncthread
def openCells(sheet, col, rows):
    for r in Progress(rows):
        vd.push(openCell(col, r))

    vd.sync()

@TableSheet.api
def openRowPyobj(sheet, rowidx):
    'Return Sheet of raw Python object of row.'
    if sheet.nRows == 0: vd.fail('no row to dive into')
    return PyobjSheet("%s[%s]" % (sheet.name, rowidx), source=sheet.rows[rowidx])

@TableSheet.api
def openCellPyobj(sheet, col, rowidx):
    'Return Sheet of raw Python object of cell.'
    if col is None or sheet.nRows == 0: vd.fail('no cell to dive into')
    name = f'{sheet.name}[{rowidx}].{col.name}'
    return PyobjSheet(name, source=col.getValue(sheet.rows[rowidx]))


@BaseSheet.api
def inputPythonExpr(sheet):
    def launch_repl(v, i):
        import code
        with SuspendCurses():
            code.InteractiveConsole(locals=locals()).interact()
        return v, i
    return vd.input("eval: ", "expr", completer=visidata.CompleteExpr(), bindings={'^X': launch_repl})

BaseSheet.addCommand('^X', 'pyobj-expr', 'expr=inputPythonExpr(); vd.push(PyobjSheet(expr, source=sheet.evalExpr(expr)))', 'evaluate Python expression and open result as Python object')
BaseSheet.addCommand('', 'exec-python', 'expr = input("exec: ", "expr", completer=CompleteExpr()); exec(expr, getGlobals(), LazyChainMap(sheet, *vd.contexts, locals=vd.getGlobals()))', 'execute Python statement with expression scope')
BaseSheet.addCommand('g^X', 'import-python', 'modname=input("import: ", type="import_python"); exec("import "+modname, getGlobals())', 'import Python module in the global scope')
BaseSheet.addCommand('z^X', 'pyobj-expr-row', 'expr = input("eval over current row: ", "expr", completer=CompleteExpr()); vd.push(PyobjSheet(expr, source=evalExpr(expr, row=cursorRow)))', 'evaluate Python expression, in context of current row, and open result as Python object')
BaseSheet.addCommand('', 'assert-expr', 'expr=inputPythonExpr(); assert sheet.evalExpr(expr), f"{expr} not true"', 'eval Python expression and assert result is truthy')
BaseSheet.addCommand('', 'assert-expr-row', 'expr=inputPythonExpr(); assert sheet.evalExpr(expr, row=cursorRow), f"{expr} not true"', 'eval Python expression in context of current row, and assert result is truthy')

Sheet.addCommand('^Y', 'pyobj-row', 'status(type(cursorRow).__name__); vd.push(openRowPyobj(cursorRowIndex))', 'open current row as Python object')
Sheet.addCommand('z^Y', 'pyobj-cell', 'status(type(cursorValue).__name__); vd.push(openCellPyobj(cursorCol, cursorRowIndex))', 'open current cell as Python object')
BaseSheet.addCommand('g^Y', 'pyobj-sheet', 'status(type(sheet).__name__); vd.push(PyobjSheet(sheet.name+"_sheet", source=sheet))', 'open current sheet as Python object')

Sheet.addCommand('', 'open-row-basic', 'vd.push(TableSheet.openRow(sheet, cursorRow))', 'dive into current row as basic table (ignoring subsheet dive)')
Sheet.addCommand(ENTER, 'open-row', 'vd.push(openRow(cursorRow)) if cursorRow else vd.fail("no row to open")', 'open current row with sheet-specific dive')
Sheet.addCommand('z'+ENTER, 'open-cell', 'vd.push(openCell(cursorCol, cursorRow))', 'open sheet with copies of rows referenced in current cell')
openRows
Sheet.addCommand('g'+ENTER, 'dive-selected', 'openRows(selectedRows)', 'open all selected rows')
Sheet.addCommand('gz'+ENTER, 'dive-selected-cells', 'openCells(cursorCol, selectedRows)', 'open all selected cells')

PyobjSheet.addCommand('v', 'visibility', 'sheet.options.visibility = 0 if sheet.options.visibility else 2; reload()', 'toggle show/hide for methods and hidden properties')
PyobjSheet.addCommand('gv', 'show-hidden', 'sheet.options.visibility = 2; reload()', 'show methods and hidden properties')
PyobjSheet.addCommand('zv', 'hide-hidden', 'sheet.options.visibility -= 1; reload()', 'hide methods and hidden properties')

vd.addGlobals({
    'PythonSheet': PythonSheet,
    'ListOfDictSheet': ListOfDictSheet,
    'SheetDict': SheetDict,
    'PyobjSheet': PyobjSheet,
    'view': view,
})

vd.addMenuItems('''
    View > Visibility > Methods and dunder attributes > show > show-hidden
    View > Visibility > Methods and dunder attributes > hide > hide-hidden
    Row > Dive into > open-row
    System > Python > import library > import-python
    System > Python > current sheet > pyobj-sheet
    System > Python > current row > pyobj-row
    System > Python > current cell > pyobj-cell
    System > Python > expression > pyobj-expr
    System > Python > exec() > exec-python
''')



import pytest
from unittest.mock import Mock, call

import visidata
from visidata import iterchars


class TestClipText:
    @pytest.mark.parametrize('s, dispw', [
        ('abcdef', 6),
        ('桜 高橋', 7),
        ('[:onclick sidebar-toggle][:reverse] b to toggle sidebar [:]', 21),
    ])
    def test_dispwidth(self, s, dispw):
        assert visidata.dispwidth(s) == dispw

    @pytest.mark.parametrize('s, w, clippeds, clippedw', [
        ('b to', 4, 'b to', 4),
        ('abcde', 8, 'abcde', 5),
        (' jsonl', 5, ' jso…', 5),
        ('abcdで', 6, 'abcdで', 6),
        ('abcdで', 5, 'abcd…', 5),
        ('a', 1, 'a', 1),
        ('ab', 1, '…', 1),
        ('abc', 2, 'a…', 2),
        ('で', 1, '…', 1),
        ('でで', 1, '…', 1),
        ('でで', 2, '…', 1),
        ('でで', 3, 'で…', 3),
        ('ででで', 4, 'で…', 3),
        ('ででで', 5, 'でで…', 5),
        ('', 1, '', 0),
        ('', None, '', 0),
        ('abcdef', None, 'abcdef', 6),
        ('ででで', None, 'ででで', 6),
        ('で'*100, None, 'で'*100, 2*100),
        (iterchars([1,2,3]), 4, '[3]…', 4),
        (iterchars([1,2,3]), 7, '[3] 1;…', 7),
        (iterchars([1,2,3]), 9, '[3] 1; 2…', 9),
        (iterchars([1,2,3]), 11, '[3] 1; 2; 3', 11),
        (iterchars([1,2,3]), 12, '[3] 1; 2; 3', 11),
        (iterchars({'a':1, 'b':2, 'c':3}), 7, '{3} a=…', 7),
        (iterchars({'a':1, 'b':2, 'c':3}), 15, '{3} a=1 b=2 c=3', 15),
        (iterchars({'a':1, 'b':2, 'で':3}), 13, '{3} a=1 b=2 …', 13),
        (iterchars({'a':1, 'b':2, 'で':3}), 16, '{3} a=1 b=2 で=3', 16),
    ])
    def test_clipstr(self, s, w, clippeds, clippedw):
        clips, clipw = visidata.clipstr(s, w)
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, w, clippeds, clippedw', [
        ('b to', 4, 'b to', 4),
        ('abcde', 8, 'abcde', 5),
        (' jsonl', 5, ' jsあ', 5),
        ('abcdで', 6, 'abcdで', 6),
        ('abcdで', 5, 'abcあ', 5),
        ('a', 1, 'a', 1),
        ('ab', 1, 'a', 1),
        ('abc', 2, 'あ', 2),
        ('で', 1, '', 0),
        ('でで', 1, '', 0),
        ('でで', 2, 'あ', 2),
        ('でで', 3, 'あ', 2),
        ('ででで', 4, 'であ', 4),
        ('ででで', 5, 'であ', 4),
        ('', 1, '', 0),
        ('', None, '', 0),
        ('abcdef', None, 'abcdef', 6),
        ('ででで', None, 'ででで', 6),
        ('で'*100, None, 'で'*100, 2*100),
        (iterchars([1,2,3]), 1, '[', 1),
        (iterchars({'a':1, 'b':2, 'c':3}), 1, '{', 1),
    ])
    def test_clipstr_wide_truncator(self, s, w, clippeds, clippedw):
        clips, clipw = visidata.clipstr(s, w, truncator='あ')
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, w, clippeds, clippedw', [
        ('b to', 4, 'b to', 4),
        ('abcde', 8, 'abcde', 5),
        (' jsonl', 5, ' json', 5),
        ('abcdで', 6, 'abcdで', 6),
        ('abcdで', 5, 'abcd', 4),
        ('a', 1, 'a', 1),
        ('ab', 1, 'a', 1),
        ('abc', 2, 'ab', 2),
        ('で', 1, '', 0),
        ('でで', 1, '', 0),
        ('でで', 2, 'で', 2),
        ('でで', 3, 'で', 2),
        ('ででで', 4, 'でで', 4),
        ('ででで', 5, 'でで', 4),
        ('', 1, '', 0),
        ('', None, '', 0),
        ('abcdef', None, 'abcdef', 6),
        ('ででで', None, 'ででで', 6),
        ('で'*100, None, 'で'*100, 2*100),
    ])
    def test_clipstr_empty_truncator(self, s, w, clippeds, clippedw):
        clips, clipw = visidata.clipstr(s, w, truncator='')
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, w, truncator, clippeds, clippedw', [
        ('first\nsecond\n\nthird\n\n\n', 22, '',  'first·second··third···', 22),
        ('first\nsecond\n\nthird\n\n\n', 22, '…', 'first·second··third···', 22),
        ('first\nsecond\n\nthird\n\n\n', 21, '',  'first·second··third··', 21),
        ('first\nsecond\n\nthird\n\n\n', 21, '…', 'first·second··third·…', 21),
        (''.join([chr(i) for i in range(256)]), 256, '',
            '································ !"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~··································¡¢£¤¥¦§¨©ª«¬\xad®¯°±²³´µ¶·¸¹º»¼½¾¿ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖ×ØÙÚÛÜÝÞßàáâãäåæçèéêëìíîïðñòóôõö÷øùúûüýþÿ', 256),
    ])
    def test_clipstr_unprintable(self, s, w, truncator, clippeds, clippedw):
        clips, clipw = visidata.clipstr(s, w, truncator=truncator, oddspace='·')
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, w, clippeds, clippedw', [
        ('b to', 4, 'b to', 4),
        ('abcde', 8, 'abcde', 5),
        (' jsonl', 5, 'jsonl', 5),
        ('abcdで', 6, 'abcdで', 6),
        ('abcdで', 5, 'bcdで', 5),
        ('でbcdで', 6, 'bcdで', 5),
        ('でbcdefghiで', 10, 'bcdefghiで', 10),
        ('でbcdefghiで', 3, 'iで', 3),
        ('でbcdで', 2, 'で', 2),
        ('でbcdで', 1, '', 0),
        ('でbcdで', 0, '', 0),
        ('でbcdで', -1, '', 0),
    ])
    def test_clipstr_start(self, s, w, clippeds, clippedw):
        clips, clipw = visidata.clipstr_start(s, w)
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, w, clippeds, clippedw', [
        ('aAbcで', 6, 'aAbcで', 6),
        ('aAbcで', 7, 'aAbcで', 6),
        ('aAbcで', 1000, 'aAbcで', 6),
        ('aAbcで', 5, '…bcで', 5),
        ('でbcで', 5, '…bcで', 5),
        ('でででででbcで', 5, '…bcで', 5),
        ('でbcで', 3, '…で', 3),
        ('でbcで', 2, '…', 1),
        ('でbcで', 1, '…', 1),
        ('でbcで', 0, '', 0),
        ('でbcで', -1, '', 0),
    ])
    def test_clipstr_start_truncator(self, s, w, clippeds, clippedw):
        clips, clipw = visidata.clipstr_start(s, w, truncator='…')
        assert clips == clippeds
        assert clipw == clippedw

    @pytest.mark.parametrize('s, w, clippeds, clippedw', [
        ('1234567890', 6, '12…890', 6),
        ('1234567890', 7, '123…890', 7),
        ('1234567890', 8, '123…7890', 8),
        ('1234567890', 9, '1234…7890', 9),
        ('1234567890', 10, '1234567890', 10),
        ('1234567890', 11, '1234567890', 10),
        ('1234567890', 99, '1234567890', 10),
        # all full-width characters
        ('ででででで', 0,  '', 0),
        ('ででででで', 1,  '…', 1),
        ('ででででで', 2,  '…', 1),
        ('ででででで', 3,  '…で', 3),
        ('ででででで', 4,  '…で', 3),
        ('ででででで', 5,  'で…で', 5),
        ('ででででで', 6,  'で…で', 5),
        ('ででででで', 7,  'で…でで', 7),
        ('ででででで', 8,  'で…でで', 7),
        ('ででででで', 9,  'でで…でで', 9),
        ('ででででで', 10, 'ででででで', 10),
        ('ででででで', 11, 'ででででで', 10),
        ('ででででで', 99, 'ででででで', 10),
        # odd string length, with mix of full-width characters
        ('ででaaでa', 0,  '', 0),
        ('ででaaでa', 1,  '…', 1),
        ('ででaaでa', 2,  '…a', 2),
        ('ででaaででa', 3,  '…a', 2),
        ('ででaaででa', 4,  '…でa', 4),
        ('ででaaででa', 5,  'で…a', 4),
        ('ででaaででa', 6,  'で…でa', 6),
    ])
    def test_clipstr_middle(self, s, w, clippeds, clippedw):
        clips, clipw = visidata.clipstr_middle(s, w)
        assert clips == clippeds
        assert clipw == clippedw

    def test_clipdraw_chunks(self):
        prechunks = [
            ('', 'x'),
            ('', 'jsonl'),
        ]
        scr = Mock()
        scr.getmaxyx.return_value = (80,25)
        visidata.clipdraw_chunks(scr, 0, 0, prechunks, visidata.ColorAttr(), w=5)
        scr.addstr.assert_has_calls([
                call(0, 0, 'x', 0),
                call(0, 1, 'jso…', 0),
        ], any_order=True)



from visidata import vd, VisiData, Sheet, asyncthread, Progress, Column

# rowdef: Column
@VisiData.api
class TransposeSheet(Sheet):
    def beforeLoad(self):
        # key rows become column names
        col = Column('_'.join(c.name for c in self.source.keyCols),
                getter=lambda c,origcol: origcol.name)
        # associate column with sheet
        col.recalc(self)
        self.columns = [col]
        self.setKeys(self.columns)

    def loader(self):
        # rows become columns
        for row in Progress(self.source.rows, 'transposing'):
            self.addColumn(Column('_'.join(map(str, self.source.rowkey(row))),
                                  getter=lambda c,origcol,row=row: origcol.getValue(row),
                                  setter=lambda c,origcol,v,row=row: origcol.setValue(row, v)))

        # columns become rows
        self.rows = list(self.source.nonKeyVisibleCols)

Sheet.addCommand('T', 'transpose', 'vd.push(TransposeSheet(name+"_T", source=sheet))', 'open new sheet with rows and columns transposed')

vd.addMenuItems('Data > Transpose > transpose')
vd.addGlobals(TransposeSheet=TransposeSheet)



from visidata import vd, BaseSheet, ENTER, colors, dispwidth
import curses


def boundingBox(rows):
    'Return (xmin, ymin, xmax, ymax) of rows.'
    xmin, ymin, xmax, ymax = 9999, 9999, 0, 0
    for r in rows:
        if r.x is not None:
            xmin = min(xmin, r.x)
            xmax = max(xmax, r.x + (r.w or dispwidth(r.text or '')))
        if r.y is not None:
            ymin = min(ymin, r.y)
            ymax = max(ymax, r.y + (r.h or 0))
    return xmin, ymin, xmax, ymax


class CharBox:
    def __init__(self, scr=None, x1=0, y1=0, w=None, h=None):
        scrh, scrw = scr.getmaxyx() if scr else (25, 80)
        self.scr = scr
        self.x1 = x1
        self.y1 = y1
        self.w = scrw if w is None else w
        self.h = scrh if h is None else h

        self.normalize()

    def __str__(self):
        return f'({self.x1}+{self.w},{self.y1}+{self.h})'

    def normalize(self):
        'Make sure w and h are non-negative, swapping coordinates as needed.'
        if self.w < 0:
            self.x1 += self.w
            self.w = -self.w

        if self.h < 0:
            self.y1 += self.h
            self.h = -self.h

    @property
    def x2(self):
        return self.x1+self.w+1

    @x2.setter
    def x2(self, v):
        self.w = v-self.x1-1
        self.normalize()

    @property
    def y2(self):
        return self.y1+self.h+1

    @y2.setter
    def y2(self, v):
        self.h = v-self.y1-1
        self.normalize()

    def contains(self, b):
        'Return True if this box contains any part of the given x,y,w,h.'
        xA = max(self.x1, b.x1)  # left
        xB = min(self.x2, b.x2)  # right
        yA = max(self.y1, b.y1)  # top
        yB = min(self.y2, b.y2)  # bottom
        return xA < xB-1 and yA < yB-1   # xA+.5 < xB-.5 and yA+.5 < yB-.5


class TextCanvas(BaseSheet):
    @property
    def rows(self):
        return self.source.rows

    @rows.setter
    def rows(self, v):
        pass

    def reload(self):
        pass

    def draw(self, scr):
        for i in range(self.cursorBox.h):
            for j in range(self.cursorBox.w):
                scr.addstr(self.cursorBox.y1+i, self.cursorBox.x1+j, ' ', colors.color_current_row)

    def commandCursor(self, execstr):
        if 'cursor' in execstr:
            return '%s %s' % (self.cursorBox.x1, self.cursorBox.x2), '%s %s' % (self.cursorBox.y1, self.cursorBox.y2)
        return '', ''

    def checkCursor(self):
        self.cursorBox.x1 = min(self.windowWidth-2, max(0, self.cursorBox.x1))
        self.cursorBox.y1 = min(self.windowHeight-2, max(0, self.cursorBox.y1))

    def iterbox(self, box, n=None):
        'Return *n* top elements from each cell within the given *box*.'
        ret = list()
        for r in self.source.rows:
            if r.pos.x is None or r.pos.y is None: continue
            if box.contains(CharBox(None, r.pos.x, r.pos.y, 1, 1)):
                ret.append(r)

        return ret[:-n] if n else ret

    def itercursor(self, n=None):
        return self.iterbox(self.cursorBox, n=n)

    @property
    def cursorRows(self):
        return list(self.iterbox(self.cursorBox))

    def slide(self, rows, dx, dy):
        maxX, maxY = self.windowWidth, self.windowHeight
        x1, y1, x2, y2 = boundingBox(rows)
        dx = -x1 if x1+dx < 0 else (maxX-x2-1 if x2+dx > maxX-1 else dx)
        dy = -y1 if y1+dy < 0 else (maxY-y2-1 if y2+dy > maxY-1 else dy)
        xcol = self.source.column('x')
        ycol = self.source.column('y')
        for r in rows:
            oldx = xcol.getValue(r)
            oldy = ycol.getValue(r)
            if oldx is not None:
                xcol.setValue(r, oldx+dx)
            if oldy is not None:
                ycol.setValue(r, oldy+dy)


TextCanvas.addCommand('', 'go-down', 'cursorBox.y1 += 1')
TextCanvas.addCommand('', 'go-up', 'cursorBox.y1 -= 1')
TextCanvas.addCommand('', 'go-left', 'cursorBox.x1 -= 1')
TextCanvas.addCommand('', 'go-right', 'cursorBox.x1 += 1')
TextCanvas.addCommand('kRIT5', 'resize-cursor-wider', 'cursorBox.w += 1', 'increase cursor width by one character')
TextCanvas.addCommand('kLFT5', 'resize-cursor-thinner', 'cursorBox.w -= 1', 'decrease cursor width by one character')
TextCanvas.addCommand('kUP5', 'resize-cursor-shorter', 'cursorBox.h -= 1', 'decrease cursor height by one character')
TextCanvas.addCommand('kDN5', 'resize-cursor-taller', 'cursorBox.h += 1', 'increase cursor height by one character')
TextCanvas.addCommand('gzKEY_LEFT', 'resize-cursor-min-width', 'cursorBox.w = 1')
TextCanvas.addCommand('gzKEY_UP', 'resize-cursor-min-height', 'cursorBox.h = 1')
TextCanvas.addCommand('z_', 'resize-cursor-min', 'cursorBox.h = cursorBox.w = 1')
TextCanvas.addCommand('g_', 'resize-cursor-max', 'cursorBox.x1=cursorBox.y1=0; cursorBox.h=maxY+1; cursorBox.w=maxX+1')
TextCanvas.bindkey('zKEY_RIGHT', 'resize-cursor-wider')
TextCanvas.bindkey('zKEY_LEFT', 'resize-cursor-thinner')
TextCanvas.bindkey('zKEY_UP', 'resize-cursor-shorter')
TextCanvas.bindkey('zKEY_DOWN', 'resize-cursor-taller')
TextCanvas.addCommand('BUTTON1_PRESSED', 'move-cursor', 'sheet.cursorBox = CharBox(None, mouseX, mouseY, 1, 1)', 'start cursor box with left mouse button press')
TextCanvas.addCommand('BUTTON1_RELEASED', 'end-cursor', 'cursorBox.x2=mouseX+2; cursorBox.y2=mouseY+2; cursorBox.normalize()', 'end cursor box with left mouse button release')

TextCanvas.addCommand('s', 'select-cursor', 'source.select(cursorRows)')
TextCanvas.addCommand('t', 'toggle-cursor', 'source.toggle(cursorRows)')
TextCanvas.addCommand('u', 'unselect-cursor', 'source.unselect(cursorRows)')
TextCanvas.addCommand('gs', 'select-all', 'source.select(source.rows)')
TextCanvas.addCommand('gt', 'toggle-all', 'source.toggle(source.rows)')
TextCanvas.addCommand('gu', 'unselect-all', 'source.clearSelected()')
TextCanvas.addCommand('zs', 'select-top-cursor', 'source.select(list(itercursor(n=1)))')
TextCanvas.addCommand('zt', 'toggle-top-cursor', 'source.toggle(list(itercursor(n=1)))')
TextCanvas.addCommand('zu', 'unselect-top-cursor', 'source.unselect(list(itercursor(n=1)))')
TextCanvas.addCommand('d', 'delete-cursor', 'source.deleteBy(lambda r,rows=cursorRows: r in rows)', 'delete first item under cursor')
TextCanvas.addCommand('gd', 'delete-selected', 'source.deleteSelected()', 'delete selected rows on source sheet')
TextCanvas.addCommand(ENTER, 'dive-cursor', 'vs=copy(source); vs.rows=cursorRows; vs.source=sheet; vd.push(vs)', 'dive into source rows under cursor')
TextCanvas.addCommand('g'+ENTER, 'dive-selected', 'vd.push(type(source)(source=sheet, rows=source.selectedRows))', 'dive into selected source rows')

TextCanvas.addCommand('H', 'slide-left-obj',       'slide(source.selectedRows, -1, 0)', 'slide selected objects left one character')
TextCanvas.addCommand('J', 'slide-down-obj',       'slide(source.selectedRows, 0, +1)', 'slide selected objects down one character')
TextCanvas.addCommand('K', 'slide-up-obj',         'slide(source.selectedRows, 0, -1)', 'slide selected objects up one character')
TextCanvas.addCommand('L', 'slide-right-obj',      'slide(source.selectedRows, +1, 0)', 'slide selected objects right one character')
TextCanvas.addCommand('gH', 'slide-leftmost-obj',  'slide(source.selectedRows, -maxX, 0)', 'slide selected objects all the way left')
TextCanvas.addCommand('gJ', 'slide-bottom-obj',    'slide(source.selectedRows, 0, +maxY)', 'slide all selected objects all the way bottom')
TextCanvas.addCommand('gK', 'slide-top-obj',       'slide(source.selectedRows, 0, -maxY)', 'slide all selected objects all the way top')
TextCanvas.addCommand('gL', 'slide-rightmost-obj', 'slide(source.selectedRows, +maxX, 0)', 'slide all selected objects all the way right')


TextCanvas.init('cursorBox', lambda: CharBox(None, 0,0,1,1))

vd.addGlobals({
    'CharBox': CharBox,
    'boundingBox': boundingBox,
    'TextCanvas': TextCanvas,
    })



from visidata import vd, VisiData, Sheet

@Sheet.api
def hint_rename_col(sheet):
    if vd.cleanName(sheet.cursorCol.name) != sheet.cursorCol.name:
        return 5, f"[:hint]The current column can't be used in an expression because [:code]{sheet.cursorCol.name}[/] is not a valid Python identifier. [:onclick rename-col]Rename the column[/] with `^`.[/]"


@VisiData.api
def addUndoColNames(vd, cols):
    oldnames = [(c, c.name) for c in cols]
    def _undo():
        for c, name in oldnames:
            c.name = name
    vd.addUndo(_undo)


@Sheet.api
def updateColNames(sheet, rows, cols, overwrite=False):
    vd.addUndoColNames(cols)
    for c in cols:
        if not c._name or overwrite:
            c.name = "\n".join(c.getDisplayValue(r) for r in rows)


Sheet.addCommand('^', 'rename-col', 'vd.addUndoColNames([cursorCol]); cursorCol.name = editCell(cursorVisibleColIndex, -1, value=cleanName(cursorCol.name))', 'rename current column')
Sheet.addCommand('z^', 'rename-col-selected', 'updateColNames(selectedRows or [cursorRow], [sheet.cursorCol], overwrite=True)', 'rename current column to combined contents of current cell in selected rows (or current row)')
Sheet.addCommand('g^', 'rename-cols-row', 'updateColNames(selectedRows or [cursorRow], sheet.visibleCols)', 'rename all unnamed visible columns to contents of selected rows (or current row)')
Sheet.addCommand('gz^', 'rename-cols-selected', 'updateColNames(selectedRows or [cursorRow], sheet.visibleCols, overwrite=True)', 'rename all visible columns to combined contents of selected rows (or current row)')


vd.addMenuItems('''
    Column > Rename > current column > rename-col
    Column > Rename > from selected cells > current column > rename-col-selected
    Column > Rename > from selected cells > unnamed columns > rename-cols-row
    Column > Rename > from selected cells > all columns > rename-cols-selected
''')



from contextlib import contextmanager
import operator
import string
import re

'Various helper classes and functions.'

__all__ = ['AlwaysDict', 'AttrDict', 'DefaultAttrDict', 'moveListItem', 'namedlist', 'classproperty', 'MissingAttrFormatter', 'getitem', 'setitem', 'getitemdef', 'getitemdeep', 'setitemdeep', 'getattrdeep', 'setattrdeep', 'ExplodingMock', 'ScopedSetattr', 'colname_letters']


class AlwaysDict(dict):
    'return same val for all keys'
    def __init__(self, val, **kwargs):
        super().__init__(**kwargs)
        self._val = val

    def __getitem__(self, k):
        return self._val

class AttrDict(dict):
    'Augment a dict with more convenient .attr syntax.  not-present keys return None.'
    def __getattr__(self, k):
        try:
            v = self[k]
            if isinstance(v, dict) and not isinstance(v, AttrDict):
                v = AttrDict(v)
            return v
        except KeyError as e:
            if k.startswith("__"):
                raise AttributeError from e
            return None

    def __setattr__(self, k, v):
        self[k] = v

    def __dir__(self):
        return self.keys()


class DefaultAttrDict(dict):
    'Augment a dict with more convenient .attr syntax.  not-present keys store new DefaultAttrDict.  like a recursive defaultdict.'
    def __getattr__(self, k):
        if k not in self:
            if k.startswith("__"):
                raise AttributeError from e
            self[k] = DefaultAttrDict()
        return self[k]

    def __setattr__(self, k, v):
        self[k] = v

    def __dir__(self):
        return self.keys()




class classproperty(property):
    def __get__(self, cls, obj):
        return classmethod(self.fget).__get__(None, obj or cls)()


def moveListItem(L, fromidx, toidx):
    "Move element within list `L` and return element's new index."
    toidx = min(max(toidx, 0), len(L)-1)
    fromidx = min(max(fromidx, 0), len(L)-1)
    r = L.pop(fromidx)
    L.insert(toidx, r)
    return toidx


def setitem(r, i, v):  # function needed for use in lambda
    r[i] = v
    return True

def getitem(o, k, default=None):
    return default if o is None else o[k]

def getitemdef(o, k, default=None):
    try:
        return default if o is None else o[k]
    except Exception:
        return default


def getattrdeep(obj, attr, *default, getter=getattr):
    try:
        'Return dotted attr (like "a.b.c") from obj, or default if any of the components are missing.'
        if not isinstance(attr, str):
            return getter(obj, attr, *default)

        try:  # if attribute exists, return toplevel value, even if dotted
            if attr in obj:
                return getter(obj, attr)
        except RecursionError:  #1696
            raise
        except Exception as e:
            pass

        attrs = attr.split('.')
        for a in attrs[:-1]:
            obj = getter(obj, a)

        return getter(obj, attrs[-1])
    except Exception as e:
        if not default: raise
        return default[0]


def setattrdeep(obj, attr, val, getter=getattr, setter=setattr):
    'Set dotted attr (like "a.b.c") on obj to val.'
    if not isinstance(attr, str):
        return setter(obj, attr, val)

    try:  # if attribute exists, overwrite toplevel value, even if dotted
        getter(obj, attr)
        return setter(obj, attr, val)
    except Exception as e:
        pass

    attrs = attr.split('.')
    for a in attrs[:-1]:
        try:
            obj = getter(obj, a)
        except Exception as e:
            obj = obj[a] = type(obj)()  # assume homogeneous nesting

    setter(obj, attrs[-1], val)


def getitemdeep(obj, k, *default):
    if not isinstance(k, str):
        try:
            return obj[k]
        except IndexError:
            pass
    return getattrdeep(obj, k, *default, getter=getitem)

def setitemdeep(obj, k, val):
    return setattrdeep(obj, k, val, getter=getitemdef, setter=setitem)


def namedlist(objname, fieldnames):
    'like namedtuple but editable'
    class NamedListTemplate(list):
        __name__ = objname
        _fields = fieldnames

        def __init__(self, L=None, **kwargs):
            if L is None:
                L = [None]*len(self._fields)
            elif len(L) < len(self._fields):
                L.extend([None]*(len(self._fields) - len(L)))
            super().__init__(L)
            for k, v in kwargs.items():
                setattr(self, k, v)

        def __getattr__(self, k):
            'to enable .fieldname'
            try:
                return self[self._fields.index(k)]
            except ValueError as e:
                raise AttributeError from e

        def __setattr__(self, k, v):
            'to enable .fieldname ='
            try:
                self[self._fields.index(k)] = v
            except ValueError:
                super().__setattr__(k, v)

    return NamedListTemplate


class ExplodingMock:
    'A mock object that raises an exception for everything except conversion to True/False.'
    def __init__(self, msg):
        self.__msg = msg

    def __getattr__(self, k):
        raise Exception(self.__msg)

    def __bool__(self):
        return False


class MissingAttrFormatter(string.Formatter):
    "formats {} fields with `''`, that would normally result in a raised KeyError or AttributeError or IndexError; intended for user customisable format strings."
    def get_value(self, key, args, kwargs):
        try:
            return super().get_value(key, args, kwargs)
        except Exception:
            raise

    def get_field(self, field_name, args, kwargs):
        try:
            return super().get_field(field_name, args, kwargs)
        except (KeyError, AttributeError, IndexError, ValueError) as e:
            return ('{' + field_name + '}', field_name)

    def format_field(self, value, format_spec):
        # value is missing
        if value is None:
            return ''
        elif not value:
            return str(value)
        return super().format_field(value, format_spec)


@contextmanager
def ScopedSetattr(obj, attrname, val):
    oldval = getattr(obj, attrname)
    try:
        setattr(obj, attrname, val)
        yield
    finally:
        setattr(obj, attrname, oldval)

def colname_letters(num):
    '''*num* is a 1-based integer: 1, 2, 3... gives A B C .. Z AA AB .. ZZ AAA .. to infinity; *num* of 0 returns the empty string'''
    # credit to https://stackoverflow.com/questions/48983939/convert-a-number-to-excel-s-base-26/48984697#48984697
    def divmod_excel(n):
        a, b = divmod(n, 26)
        if b == 0:
            return a - 1, b + 26
        return a, b
    chars = []
    while num > 0:
        num, d = divmod_excel(num)
        chars.append('-ABCDEFGHIJKLMNOPQRSTUVWXYZ'[d])
    return ''.join(reversed(chars))



from visidata import vd, VisiData, Sheet, Column, AttrColumn, date, vlen, asyncthread, Path, namedlist, PyobjSheet, modtime, AttrDict

from .gitsheet import GitSheet

@VisiData.api
def guess_git(vd, p):
    if (p/'.git').is_dir():
        return dict(filetype='git', _likelihood=10)


@VisiData.api
def open_git(vd, p):
    return vd.git_status(p, [])


@VisiData.api
def git_repos(vd, p, args):
    return GitRepos(p.base_stem, source=p)


class GitLinesColumn(Column):
    def __init__(self, name, cmd, *args, **kwargs):
        super().__init__(name, cache='async', **kwargs)
        cmdparts = cmd.split()
        if cmdparts[0] == 'git':
            cmdparts = cmdparts[1:]
        self.gitargs = cmdparts + list(args)

    def calcValue(self, r):
        lines = list(GitSheet(source=r).git_lines(*self.gitargs))
        if lines:
            return lines


class GitAllColumn(GitLinesColumn):
    def calcValue(self, r):
        return GitSheet(source=r).git_all(*self.gitargs).strip()



class GitRepos(GitSheet):
    guide = '''
        # git repos
        A list of git repositories under `{sheet.source}`

        - `Enter` to open the status sheet for the current repo
    '''
    rowtype = 'git repos'  # rowdef: Path
    columns = [
        Column('repo', type=str, width=30),
        GitAllColumn('branch', 'git rev-parse --abbrev-ref HEAD', width=8),
        GitLinesColumn('diffs', 'git diff --no-color', type=vlen, width=8),
        GitLinesColumn('staged_diffs', 'git diff --cached', type=vlen, width=8),
        GitLinesColumn('branches', 'git branch --no-color', type=vlen, width=10),
        GitLinesColumn('stashes', 'git stash list', type=vlen, width=8),
        Column('modtime', type=date, getter=lambda c,r: modtime(r)),
    ]
    nKeys = 1

    def iterload(self):
        import glob
        for fn in glob.glob('**/.git', root_dir=self.source, recursive=True):
            yield Path(fn).parent


    def openRow(self, row):
        return vd.git_status(row, [])

    def openCell(self, col, row):
        val = col.getValue(row)
        return PyobjSheet(getattr(val, '__name__', ''), source=val)



from visidata import VisiData, vd, options, Progress

def markdown_link(s, href):
    if not href:
        return s

    return f'[{s}]({href})'

def markdown_escape(s, style='orgmode'):
    if style == 'jira':
        return s

    ret = ''
    for ch in s:
        if ch in '\\`*_{}[]()>#+-.!':
            ret += '\\'+ch
        else:
            ret += ch
    return ret

def markdown_colhdr(col):
    if vd.isNumeric(col):
        return ('-' * (col.width-1)) + ':'
    else:
        return '-' * (col.width or options.default_width)

def write_md(p, *vsheets, md_style='orgmode'):
    'pipe tables compatible with org-mode'

    if md_style == 'jira':
        delim = '||'
    else:
        delim = '|'

    with p.open(mode='w', encoding=vsheets[0].options.save_encoding) as fp:
        for vs in vsheets:
            if len(vsheets) > 1:
                fp.write('# %s\n\n' % vs.name)

            hdrs = []
            for col in vs.visibleCols:
                if col.name.endswith('_href'):
                    continue
                hdrs.append('%-*s' % (col.width or options.default_width, markdown_escape(col.name, md_style)))

            fp.write(delim + delim.join(hdrs) + delim + '\n')

            if md_style == 'orgmode':
                fp.write('|' + '|'.join(markdown_colhdr(col) for col in vs.visibleCols if not col.name.endswith('_href')) + '|\n')

            with Progress(gerund='saving'):
                for dispvals in vs.iterdispvals(format=True):
                    vals = []
                    for col, val in dispvals.items():
                        if col.name.endswith('_href'):
                            continue
                        val = markdown_escape(val, md_style)
                        linkcol = vs.colsByName.get(col.name + '_href')
                        if linkcol:
                            val = markdown_link(val, dispvals.get(linkcol))
                        vals.append('%-*s' % (col.width or options.default_width, val))
                    fp.write('|' + '|'.join(vals) + '|\n')

            fp.write('\n')


@VisiData.api
def save_md(vd, p, *sheets):
    write_md(p, *sheets, md_style='orgmode')


@VisiData.api
def save_jira(vd, p, *sheets):
    write_md(p, *sheets, md_style='jira')



import os
import time

from visidata import vd, BaseSheet, Sheet, asyncignore, asyncthread, asyncsingle_queue, Path, ScopedSetattr


@BaseSheet.api
@asyncignore
def reload_every(sheet, seconds:int):
    while True:
        # continue reloading till vd.remove() runs, like when the sheet is quit
        if not sheet in vd.sheets:
            break
        vd.sync(sheet.reload())
        time.sleep(seconds)


@BaseSheet.api
@asyncthread
def reload_modified(sheet):
    'Spawn thread to call sheet.reload_rows when sheet.source mtime has changed.'
    p = sheet.source
    assert isinstance(p, Path)
    assert not p.is_url()

    mtime = 0
    while True:
        t = os.stat(p).st_mtime
        if t != mtime:
            mtime = t
            sheet.reload_rows()
        time.sleep(1)


@Sheet.api
@asyncsingle_queue
def reload_rows(self):
    '''Reload rows from ``self.source``, keeping current columns intact.  Async. If previous calls are running, waits for them to finish.'''
    with (ScopedSetattr(self, 'loading', True),
          ScopedSetattr(self, 'checkCursor', lambda: True),
          ScopedSetattr(self, 'cursorRowIndex', self.cursorRowIndex)):
            self.beforeLoad()
            try:
                self.loader()
                vd.status("finished loading rows")
            finally:
                self.afterLoad()


BaseSheet.addCommand('', 'reload-every', 'sheet.reload_every(input("reload interval (sec): ", value=1))', 'schedule sheet reload every N seconds') #683
BaseSheet.addCommand('', 'reload-modified', 'sheet.reload_modified()', 'reload sheet when source file modified (tail-like behavior)')  #1686
Sheet.addCommand('z^R', 'reload-rows', 'preloadHook(); reload_rows(); status("reloaded")', 'Reload current sheet leaving current columns intact')

vd.addMenuItems('''
    File > Reload > rows only > reload-rows
    File > Reload > every N seconds > reload-every
    File > Reload > when source modified > reload-modified
''')



import io
import json
import socket
from unittest.mock import Mock

from visidata import vd, VisiData, asyncthread, asyncignore, CommandLogRow, Sheet


vd.option('server_addr', '127.0.0.1', 'IP address to listen for commands', sheettype=None, replay=False)
vd.option('server_port', 0, 'port to listen for commands', sheettype=None, replay=False)


class SocketIO(io.RawIOBase):
    def __init__(self, sock):
        self.sock = sock

    def read(self, sz=-1):
        if (sz == -1): sz=0x7FFFFFFF
        return self.sock.recv(sz)

    def seekable(self):
        return False


@VisiData.before
def mainloop(vd, scr):
    port = vd.options.server_port
    if port:
        vd.command_listener(vd.options.server_addr, port)


@VisiData.api
@asyncignore
def command_listener(vd, addr, port):
    while True:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind((addr, port))
        s.listen(1)

        conn, (addr, inport) = s.accept()
        vd.status(f'Connection from {addr}:{inport}')
        vd.queueCommand('no-op')  # update screen
        vd.command_server(conn)


@VisiData.api
@asyncignore
def command_server(vd, conn):
    for line in SocketIO(conn):
        line = line.decode().strip()
        if line == 'draw':
            r = '\n'.join(json.dumps(d) for d in vd.sheet.capture_draw_object())
            conn.send(r.encode('utf-8')+b'\n')
        elif line.startswith('{'):
            cmd = json.loads(line)
            vd.queueCommand(**cmd)
        else:
            longname, *rest = line.split(' ', maxsplit=1)
            cmd = dict(longname=longname, input=rest[0] if rest else '')
            vd.queueCommand(**cmd)

    conn.close()


@Sheet.api
def capture_draw_object(sheet, topRowIndex=0, nScreenRows=25):
    'capture interface at the object level'
    isNull = sheet.isNullFunc()
    sortkeys = {col:rev for col, rev in sheet._ordering}
    rows = sheet.rows[topRowIndex:min(topRowIndex+nScreenRows+1, sheet.nRows)]

    for vcolidx, col in enumerate(sheet.visibleCols):
        colstate = col.__getstate__()

        if col in sortkeys:
            colstate['sort'] = 'desc' if sortkeys.get(col) else 'asc'

        yield dict(i=vcolidx, _type='column', **colstate)

    for rowidx, row in enumerate(rows):
        rowstate = dict()

        for notefunc in vd.rowNoters:
            ch = notefunc(sheet, row)
            if ch:
                rowstate['note'] = rowstate.get('note', '') + ch

        for vcolidx, col in enumerate(sheet.visibleCols):
            cellval = col.getCell(row)

            disp = ''.join(x for _, x in col.display(cellval))
            cellstate = dict(display=disp)
            notes = getattr(cellval, 'notes', '')
            try:
                if isNull and isNull(cellval.value):
                    notes += sheet.options.disp_note_none
            except (TypeError, ValueError):
                pass

            if notes:
                cellstate['notes'] = notes
            rowstate[str(vcolidx)] = cellstate

        yield dict(_type='row', i=rowidx, **rowstate)



from visidata import VisiData, Sheet, ItemColumn, date, Column


@VisiData.api
def open_mbox(vd, p):
    return MboxSheet(p.base_stem, source=p, format='mbox')
VisiData.open_mailbox = VisiData.open_mbox

@VisiData.api
def open_maildir(vd, p):
    return MboxSheet(p.base_stem, source=p, format='Maildir')

@VisiData.api
def open_mmdf(vd, p):
    return MboxSheet(p.base_stem, source=p, format='MMDF')

@VisiData.api
def open_babyl(vd, p):
    return MboxSheet(p.base_stem, source=p, format='Babyl')

@VisiData.api
def open_mh(vd, p):
    return MboxSheet(p.base_stem, source=p, format='MH')


class MboxSheet(Sheet):
    columns = [
        ItemColumn('Date', type=date),
        ItemColumn('From'),
        ItemColumn('To'),
        ItemColumn('Cc'),
        ItemColumn('Subject'),
        Column('Payload', getter=lambda c,r: r.get_payload(decode=True),
                          setter=lambda c,r,v: r.set_payload(v)),
    ]
    def iterload(self):
        import mailbox
        cls = getattr(mailbox, self.format)
        self.mailbox = cls(str(self.source), create=False)
        for r in self.mailbox.itervalues():
            yield r



import itertools
import re
from visidata import vd, Sheet, AttrDict, dispwidth


@Sheet.api
def nextColRegex(sheet, colregex):
    'Go to first visible column after the cursor matching `colregex`.'
    pivot = sheet.cursorVisibleColIndex
    for i in itertools.chain(range(pivot+1, len(sheet.visibleCols)), range(0, pivot+1)):
        c = sheet.visibleCols[i]
        if re.search(colregex, c.name, sheet.regex_flags()):
            return i

    vd.fail('no column name matches /%s/' % colregex)

@Sheet.api
def nextColName(sheet, show_cells=True):
    if len(sheet.visibleCols) == 0: vd.fail('no columns to choose from')

    prompt = 'choose column name: '
    colnames = []
    for c in sheet.visibleCols:
        dv = None
        if show_cells and len(sheet.rows) > 0:
            dv = c.getDisplayValue(sheet.cursorRow)
        #the underscore that starts _cursor_cell excludes it from being fuzzy matched
        item = AttrDict(name=c.name,
                        _cursor_cell=dv)
        colnames.append(item)

    def _fmt_colname(match, row, trigger_key):
        name = match.formatted.get('name', row.name) if match else row.name
        r = ' '*(dispwidth(prompt)-3)
        r += f'[:keystrokes]{trigger_key}[/]  '
        if show_cells and len(sheet.rows) > 0:
            # pad the right side with spaces
            # use row.name, because name from match contains
            # extra formatting characters that change its length
            n_spaces = max(20 - dispwidth(row.name), 0)
            r += name + n_spaces*' '
            r += '   '
            #todo:  does not show disp_note_none for None
            r += row._cursor_cell
        else:
            r += name
        return r

    name = vd.activeSheet.inputPalette(prompt,
            colnames,
            value_key='name',
            type='column_name',
            formatter=_fmt_colname)

    pivot = sheet.cursorVisibleColIndex
    for i in itertools.chain(range(pivot+1, len(sheet.visibleCols)), range(0, pivot+1)):
        if name == sheet.visibleCols[i].name:
            return i

    vd.warning(f'found no column with name: {name}')


Sheet.addCommand('c', 'go-col-regex', 'sheet.cursorVisibleColIndex=nextColRegex(inputRegex("column name regex: ", type="regex-col", defaultLast=True))', 'go to next column with name matching regex')
Sheet.addCommand('zc', 'go-col-number', 'sheet.cursorVisibleColIndex = int(input("move to column number: "))', 'go to given column number (0-based)')
Sheet.addCommand('', 'go-col-name', 'sheet.cursorVisibleColIndex=nextColName()', 'go to next column with name matching string, case-insensitive')

vd.addMenuItems('''
    Column > Goto > by regex > go-col-regex
    Column > Goto > by number > go-col-number
    Column > Goto > by name > go-col-name
''')



'Switch between packaged themes (colors and display characters)'

from visidata import vd, VisiData, Sheet, BaseSheet


vd.option('theme', '', 'display/color theme to use')

vd.themes = {}


@VisiData.before
def run(vd, *args, **kwargs):
    t = vd.options.theme
    if t:
        vd.set_theme(t)


@Sheet.api
@VisiData.api
def set_theme(obj, theme=''):
    if theme and theme not in vd.themes:
        vd.warning(f'no "{theme}" theme available')
        return

    # unset everything first
    for k in vd.options.keys():
        if k.startswith(tuple('color_ disp_ note_'.split())):
            obj.options.unset(k)

    if not theme:
        return

    if isinstance(theme, str):
        theme = vd.themes[theme]

    for k, v in theme.items():
        obj.options[k] = v


BaseSheet.addCommand('', 'theme-input', 'vd.set_theme(chooseOne([dict(key=k) for k in themes.keys()], type="theme"))', 'choose from available themes')
BaseSheet.addCommand('', 'theme-default', 'vd.set_theme()', 'reset theme to VisiData defaults')

vd.addMenuItem('View', 'Set theme', 'choose', 'theme-input')
vd.addMenuItem('View', 'Set theme', 'default', 'theme-default')



import visidata
from visidata import vd, VisiData, Sheet, IndexSheet, SequenceSheet, ColumnItem, Path, AttrDict, ColumnAttr, asyncthread, Progress, ColumnExpr, date

from .gsheets import GSheetsIndex

@VisiData.api
def open_gdrive(vd, p):
    return GDriveSheet(p.base_stem)


FILES_FIELDS_VISIBLE='''name size modifiedTime mimeType description'''.split()

FILES_FIELDS='''
id name size modifiedTime mimeType description owners
starred properties spaces version webContentLink webViewLink sharingUser lastModifyingUser shared
ownedByMe originalFilename md5Checksum size quotaBytesUsed headRevisionId imageMediaMetadata videoMediaMetadata parents
exportLinks contentRestrictions contentHints trashed
'''.split()


@VisiData.cached_property
def _drivebuild(vd):
    return vd.google_discovery.build("drive", "v3", credentials=vd.google_auth('drive.readonly'))

@VisiData.cached_property
def _gdrive(self):
    return vd.google_discovery.build("drive", "v3", credentials=vd.google_auth('drive.readonly')).files()

@VisiData.cached_property
def _gdrive_rw(self):
    return vd.google_discovery.build("drive", "v3", credentials=vd.google_auth('drive')).files()


class GDriveSheet(Sheet):
    rowtype='files'  # rowdef: AttrDict of result from Google Drive files.list API
    defer=True
    columns = [
        ColumnItem('name'),
        ColumnItem('size', type=int),
        ColumnItem('modifiedTime', type=date),
        ColumnItem('mimeType'),
        ColumnItem('name'),
        ColumnExpr('owner', expr='owners[0]["displayName"]')
    ] + [
        ColumnItem(x, width=0) for x in FILES_FIELDS if x not in FILES_FIELDS_VISIBLE
    ]

    def iterload(self):
        self.results = []
        page_token = None
        while True:
            ret = vd._gdrive.list(
                    pageSize=1000,
                    pageToken=page_token,
                    fields="nextPageToken, files(%s)" % ','.join(FILES_FIELDS)
                    ).execute()

            self.results.append(ret)

            for r in ret.get('files', []):
                yield AttrDict(r)

            page_token = ret.get('nextPageToken', None)
            if not page_token:
                break

    def openRow(self, r):
        if r.mimeType == 'application/vnd.google-apps.spreadsheet':
            return GSheetsIndex(r.name, source=r.id)
        if r.mimeType.startswith('image'):
            return vd.launchBrowser(r.webViewLink)
        return vd.openSource(r.webContentLink)

    @asyncthread
    def deleteFile(self, **kwargs):
        with Progress(total=1) as prog:
            vd._gdrive_rw.delete(**kwargs).execute()
            prog.addProgress(1)

    @asyncthread
    def putChanges(self):
        adds, mods, dels = self.getDeferredChanges()

        for row in Progress(dels.values()):
            self.deleteFile(fileId=row.id)

        vd.sync()
        self.preloadHook()
        self.reload()



# contributed by Ram Rachum (@cool-RR) via ChatGPT  #2751

from visidata import GraphSheet

@GraphSheet.api
def zoom_all_y(sheet):
    """Find the lowest and highest y values in the current visible plot and set the y range to that."""
    ymin, ymax = None, None
    
    # Find the min and max y values of all points within the current x range
    xmin, xmax = sheet.visibleBox.xmin, sheet.visibleBox.xmax
    
    for vertexes, attr, row in sheet.polylines:
        if attr in sheet.hiddenAttrs:
            continue
            
        for x, y in vertexes:
            # Check if the point is within the current x range
            if xmin <= x <= xmax:
                if ymin is None or y < ymin:
                    ymin = y
                if ymax is None or y > ymax:
                    ymax = y
    
    if ymin is None or ymax is None:
        # No visible points found
        return
    
    # Add a 5% margin on both top and bottom
    y_range = ymax - ymin
    margin = sheet.ycols[0].type(y_range * 0.05)
    
    # Calculate adjusted min/max with margin
    adj_ymin = ymin - margin
    adj_ymax = ymax + margin
    
    # Create 5 equally spaced y ticks from real min to real max (not adjusted with margin)
    step = (ymax - ymin) / 4  # 5 ticks means 4 intervals
    y_ticks = tuple(ymin + step * i for i in range(5))
    
    # Set the y range with margin
    sheet.set_y(f"{adj_ymin} {adj_ymax}")
    
    # Set custom y ticks (using the real min/max, not the adjusted ones)
    sheet.forced_y_ticks = y_ticks

GraphSheet.addCommand('g_', 'zoom-all-y', 'zoom_all_y()', 'Zoom y-axis to fit all visible data points')



from visidata import vd, VisiData, ItemColumn, RowColorizer, AttrDict, Column

from .gitsheet import GitSheet

vd.option('git_diff_algo', 'minimal', 'algorithm to use for git diff')
vd.theme_option('color_git_hunk_add', 'green', 'color for added hunk lines')
vd.theme_option('color_git_hunk_del', 'red', 'color for deleted hunk lines')
vd.theme_option('color_git_hunk_diff', 'yellow', 'color for hunk diffs')

@VisiData.api
def git_diff(vd, p, args):
    return GitDiffSheet('git-diff', source=p, gitargs=args)

def _parseStartCount(s):
    sc = s.split(',')
    if len(sc) == 2:
        return sc
    if len(sc) == 1:
        return sc[0], 1


class GitDiffSheet(GitSheet):
    columns = [
        ItemColumn('a_fn', width=0),
        ItemColumn('fn', 'b_fn', width=30, hoffset=-28),
        ItemColumn('a_lineno', type=int, width=0),
        ItemColumn('lineno', 'b_lineno', type=int, width=8),
        Column('count', width=10, getter=lambda c,r: c.sheet.hunkCount(r)),
        ItemColumn('context'),
        ItemColumn('lines', type=''.join),
    ]

    guide = '''# {sheet.cursorRow.a_fn}
{sheet.cursorLines}'''

    def hunkCount(self, row):
        return f'-{row.a_count}/+{row.b_count}'

    @property
    def cursorLines(self):
        r = ''
        for line in self.cursorRow.lines[2:]:
            if line.startswith('-'):
                line = '[:git_hunk_del]' + line + '[/]'
            elif line.startswith('+'):
                line = '[:git_hunk_add]' + line + '[/]'

            r += line + '\n'
        r = r[4:]
        return r

    def iterload(self):
        current_hunk = None

        for line in self.git_lines('diff --patch --inter-hunk-context=2 --find-renames --no-color --no-prefix', *self.gitargs):
            if line.startswith('diff'):
                diff_started = True
                continue
            if not diff_started:
                continue

            if line.startswith('---'):
                hunk_lines = [line]  # new file
                leftfn = line[4:]
            elif line.startswith('+++'):
                hunk_lines.append(line)
                rightfn = line[4:]
            elif line.startswith('@@'):
                hunk_lines.append(line)
                _, linenums, context = line.split('@@')
                leftlinenums, rightlinenums = linenums.split()
                leftstart, leftcount = _parseStartCount(leftlinenums[1:])
                rightstart, rightcount = _parseStartCount(rightlinenums[1:])
                current_hunk = AttrDict(
                    a_fn=leftfn,
                    b_fn=rightfn,
                    context=context,
                    a_lineno=int(leftstart),
                    a_count=0,
                    b_lineno=int(rightstart),
                    b_count=0,
                    lines=hunk_lines
                )
                yield current_hunk
                hunk_lines = hunk_lines[:2]  # keep file context only

            elif line[0] in ' +-':
                current_hunk.lines.append(line)
                if line[0] == '+':
                    current_hunk.a_count += 1
                elif line[0] == '-':
                    current_hunk.b_count += 1

    def openRow(self, row):
        return HunkViewer(f'{row.a_fn}:{row.a_lineno}', source=self.source, hunks=[row])


class HunkViewer(GitSheet):
    colorizers = [
        RowColorizer(4, 'color_git_hunk_add', lambda s,c,r,v: r and r.old != r.new and r.old is None),
        RowColorizer(4, 'color_git_hunk_del', lambda s,c,r,v: r and r.old != r.new and r.new is None),
        RowColorizer(5, 'color_git_hunk_diff', lambda s,c,r,v: r and r.old != r.new and r.new is not None and r.old is not None),
    ]
    columns = [
        ItemColumn('1', 'old', width=40),
        ItemColumn('2', 'new', width=40),
    ]

    def draw(self, scr):
        self.column('1').width=self.windowWidth//2-1
        self.column('2').width=self.windowWidth//2-1
        super().draw(scr)

    def iterload(self):
        nextDelIdx = None
        for hunk in self.hunks:
          for line in hunk.lines[3:]:  # diff without the patch headers
            typech = line[0]
            line = line[1:]
            if typech == '-':  # deleted
                yield AttrDict(hunk=hunk, type=typech, old=line)
                if nextDelIdx is None:
                    nextDelIdx = len(self.rows)-1
            elif typech == '+':  # added
                if nextDelIdx is not None:
                    if nextDelIdx < len(self.rows):
                        self.rows[nextDelIdx].new = line
                        nextDelIdx += 1
                        continue

                yield AttrDict(hunk=hunk, type=typech, new=line)
                nextDelIdx = None
            elif typech == ' ':  # unchanged
                yield AttrDict(hunk=hunk, type=typech, old=line, new=line)
                nextDelIdx = None
            else:
                continue  # header


HunkViewer.addCommand('2', 'git-apply-hunk', 'source.git_apply(cursorRow.hunk, "--cached"); reload()', 'apply this hunk to the index and move to the next hunk')
HunkViewer.addCommand('1', 'git-remove-hunk', 'source.git_apply(cursorRow.hunk, "--reverse"); reload()', 'remove this hunk from staging')
HunkViewer.addCommand('Enter', 'git-skip-hunk', 'hunks.pop(0); reload()', 'move to the next hunk without applying this hunk')
HunkViewer.addCommand('d', 'delete-line', 'source[7].pop(cursorRow[3]); reload()', 'delete a line from the patch')

#HunksSheet.addCommand('g^J', 'git-diff-selected', 'vd.push(HunkViewer(selectedRows or rows, source=sheet))', 'view the diffs for the selected hunks (or all hunks)')

@GitDiffSheet.api
def git_apply(sheet, row, *args):
    sheet.git('apply -p0 -', *args, _in='\n'.join(row.lines)+'\n')

    c = sheet.hunkCount(row)
    vd.status(f'applied hunk ({c})')
    sheet.reload()


#DiffSheet.addCommand('[', '', 'cursorRowIndex = findDiffRow(cursorCol.refnum, cursorRowIndex, -1)', 'go to previous diff')
#DiffSheet.addCommand(']', '', 'cursorRowIndex = findDiffRow(cursorCol.refnum, cursorRowIndex, +1)', 'go to next diff')

GitDiffSheet.addCommand('a', 'git-add-hunk', 'git_apply(cursorRow, "--cached")', 'apply this hunk to the index')

vd.addMenuItems('''
    Git > Stage > current hunk > git-add-hunk
    Git > Stage > current hunk > git-add-hunk
''')

vd.addGlobals(
    GitDiffSheet=GitDiffSheet,
    HunkViewer=HunkViewer
)



from itertools import chain

from visidata import VisiData, Progress, JsonSheet, vd


@VisiData.api
def open_yml(vd, p):
    return YamlSheet(p.base_stem, source=p)

VisiData.open_yaml = VisiData.open_yml

class YamlSheet(JsonSheet):
    def iterload(self):
        yaml = vd.importExternal('yaml', 'PyYAML')

        class PrettySafeLoader(yaml.SafeLoader):
            def construct_python_tuple(self, node):
                return tuple(self.construct_sequence(node))

        PrettySafeLoader.add_constructor(
            u'tag:yaml.org,2002:python/tuple',
            PrettySafeLoader.construct_python_tuple
        )

        with self.source.open() as fp:
            documents = yaml.load_all(fp, PrettySafeLoader)

            self.columns = []
            self._knownKeys.clear()

            # Peek at the document stream to determine how to best DWIM.
            #
            # This code is a bit verbose because it avoids slurping the generator
            # all at once into memory.
            try:
                first = next(documents)
            except StopIteration:
                # Empty file‽
                yield None
                return

            try:
                second = next(documents)
            except StopIteration:
                if isinstance(first, list):
                    # A file with a single YAML list: yield one row per list item.
                    yield from Progress(first)
                else:
                    # A file with a single YAML non-list value, e.g a dict.
                    yield first
            else:
                # A file containing multiple YAML documents: yield one row per document.
                yield from Progress(chain([first, second], documents), total=0)



import io

from visidata import VisiData, vd, options, TableSheet, ColumnItem, IndexSheet
from visidata.loaders._pandas import PandasSheet

vd.option('pdf_tables', False, 'parse PDF for tables instead of pages of text', replay=True)

@VisiData.api
def open_pdf(vd, p):
    if vd.options.pdf_tables:
        return TabulaSheet(p.base_stem, source=p)
    return PdfMinerSheet(p.base_stem, source=p)


class PdfMinerSheet(TableSheet):
    rowtype='pages' # rowdef: [pdfminer.LTPage, pageid, text]
    columns=[
        ColumnItem('pdfpage', 0, width=0),
        ColumnItem('pagenum', 1, type=int),
        ColumnItem('contents', 2),
    ]
    def iterload(self):
        vd.importExternal('pdfminer.high_level', 'pdfminer.six')
        import pdfminer.high_level
        from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
        from pdfminer.converter import TextConverter, PDFPageAggregator
        from pdfminer.layout import LAParams
        from pdfminer.pdfpage import PDFPage

        with self.source.open_bytes() as fp:
            for page in PDFPage.get_pages(fp):
                with io.StringIO() as output_string:
                    newrsrcmgr = PDFResourceManager()
                    txtconv = TextConverter(newrsrcmgr, output_string, codec=options.encoding, laparams=LAParams())
                    interpreter = PDFPageInterpreter(newrsrcmgr, txtconv)
                    interpreter.process_page(page)
                    yield [page, page.pageid, output_string.getvalue()]


class TabulaSheet(IndexSheet):
    def iterload(self):
        tabula = vd.importExternal('tabula')
        for i, t in enumerate(tabula.read_pdf(self.source, pages='all', multiple_tables=True)):
            yield PandasSheet(self.source.name, i, source=t)



import textwrap
import re

from visidata import vd, BaseSheet, options, Sheet, ColumnItem, asyncthread
from visidata import Column, vlen
from visidata import globalCommand, VisiData
import visidata


vd.option('wrap', False, 'wrap text to fit window width on TextSheet')
vd.option('save_filetype', 'tsv', 'specify default file type to save as', replay=True)


## text viewer
# rowdef: (linenum, str)
class TextSheet(Sheet):
    'Displays any iterable source, with linewrap if ``options.wrap`` is set.'
    rowtype = 'lines'  # rowdef: [linenum, text]
    filetype = 'txt'
    columns = [
        ColumnItem('linenum', 0, type=int, width=0),
        ColumnItem('text', 1),
    ]

    def iterload(self):
        yield from self.readlines(self.source)

    def readlines(self, source):
        winWidth = min(self.columns[1].width or 78, self.windowWidth-2)
        wrap = self.options.wrap
        for startingLine, text in enumerate(source):
            if wrap and text:
                for i, L in enumerate(textwrap.wrap(str(text), width=winWidth)):
                    yield [startingLine+i+1, L]
            else:
                yield [startingLine+1, text]

    def sysopen(sheet, linenum=0):
        @asyncthread
        def writelines(sheet, fn):
            with open(fn, 'w') as fp:
                for row in sheet.rows:
                    fp.write(row[1])
                    fp.write('\n')

        import tempfile
        with tempfile.NamedTemporaryFile() as temp:
            temp.close()  #2118
            writelines(sheet, temp.name)
            vd.launchEditor(temp.name, '+%s' % linenum)
            sheet.rows = []
            for r in sheet.readlines(visidata.Path(temp.name)):
                sheet.addRow(r)


# .source is list of source text lines to 'load'
# .sourceSheet is Sheet error came from
class ErrorSheet(TextSheet):
    columns = [
        ColumnItem('linenum', 0, type=int, width=0),
        ColumnItem('error', 1),
    ]
    guide = '''
        # Error Sheet
This is the traceback for an error.
- move cursor then {help.commands.sysopen_error}
- `q` to quit this error sheet.
    '''
    precious = False

    def sysopen_error(self, col, row):
        '''Open an external editor for the file relevant to the cursor line
        in the Error Sheet. If the cursor is on the first line, use the file
        mentioned at the end of the stack trace'''
        if self.rows and self.cursorRowIndex == 0:
            searchidx = len(self.rows) - 1
        else:
            searchidx = self.cursorRowIndex
        pat = re.compile(r'^ +File "(.*)", line (\d+), in ')
        for _, text in self.rows[searchidx::-1]: # rowdef: [linenum, text]
            match = pat.search(text)
            if match:
                vd.launchEditor(match.group(1), f'+{match.group(2)}')
                return

    def reload(self):
        src = self.source or (vd.lastErrors[-1] if vd.lastErrors else [])
        self.rows = list(enumerate(src))

class ErrorCellSheet(ErrorSheet):
    columns = [
        ColumnItem('linenum', 0, type=int, width=0),
        ColumnItem('cell_error', 1),
    ]
    guide = '''# Error Cell Sheet
This sheet shows the error that occurred when calculating a cell.
- move cursor then {help.commands.sysopen_error}
- `q` to quit this error sheet.
'''


class ErrorsSheet(Sheet):
    columns = [
        Column('nlines', type=vlen),
        ColumnItem('lastline', -1)
    ]
    def reload(self):
        self.rows = self.source or vd.lastErrors

    def openRow(self, row):
        return ErrorSheet(source=self.cursorRow)

@VisiData.lazy_property
def allErrorsSheet(self):
    return ErrorsSheet("errors_all")

@VisiData.lazy_property
def recentErrorsSheet(self):
    return ErrorSheet("errors_recent")



BaseSheet.addCommand('^E', 'error-recent', 'recentErrorsSheet.reload(); vd.push(recentErrorsSheet) if vd.lastErrors else status("no error")', 'view traceback for most recent error')
BaseSheet.addCommand('g^E', 'errors-all', 'vd.push(vd.allErrorsSheet)', 'view traceback for most recent errors')

Sheet.addCommand('z^E', 'error-cell', 'vd.push(ErrorCellSheet(sheet.name+"_cell_error", sourceSheet=sheet, source=getattr(cursorCell, "error", None) or fail("no error this cell")))', 'view traceback for error in current cell')

TextSheet.addCommand('^O', 'sysopen-sheet', 'sheet.sysopen(sheet.cursorRowIndex)', 'open copy of text sheet in $EDITOR and reload on exit')

ErrorSheet.addCommand('Enter', 'sysopen-error', 'sysopen_error(cursorCol, cursorRow)', 'open traceback line in $EDITOR')

TextSheet.options.save_filetype = 'txt'

vd.addGlobals({'TextSheet': TextSheet, 'ErrorSheet': ErrorSheet, 'ErrorCellSheet': ErrorCellSheet})

vd.addMenuItems('''
    View > Errors > recent > error-recent
    View > Errors > all > errors-all
    View > Errors > in cell > error-cell
''')



''' Plugin for viewing files with appropriate mailcap-specified application.
Add mailcap-view and mailcap-view-selected commands to DirSheet.

mimetype can be given explicitly with `mimetype` option; will be guessed by filename otherwise.

Usage:
   - add `import experimental.mailcap_view` to .visidatarc
   - on the DirSheet, `Ctrl+V` or `gCtrl+V` to view file(s) using mailcap entry for the guessed (or given via options) mimetype
'''

import os
from visidata import vd, DirSheet, SuspendCurses

vd.option('mailcap_mimetype', '', 'force mimetype for sysopen-mailcap')
vd.optalias('mimetype', 'mailcap_mimetype')


@DirSheet.api
def run_mailcap(sheet, p, key='view'):
    import mailcap
    import mimetypes

    mimetype = sheet.options.mailcap_mimetype
    if not mimetype:
        mimetype, encoding = mimetypes.guess_type(str(p))

    if not mimetype:
        vd.fail('no mimetype given and no guess')

    caps = mailcap.getcaps()

    plist = [f'{k}={v}' for k, v in sheet.options.getall('mailcap_').items() if k != 'mailcap_mimetype']
    cmdline, mcap_entry = mailcap.findmatch(caps, mimetype, key=key, filename=str(p), plist=plist)

    with SuspendCurses():
        os.system(cmdline)


DirSheet.addCommand('', 'sysopen-mailcap', 'run_mailcap(cursorRow)', 'open using mailcap entry for current row, guessing mimetype')
DirSheet.addCommand('', 'sysopen-mailcap-selected', 'for r in selectedRows: run_mailcap(r)', 'open selected files in succession, using mailcap')


vd.addMenuItems('''
    File > Open > using mailcap > file at cursor > sysopen-mailcap
    File > Open > using mailcap > selected files > sysopen-mailcap-selected
''')



import logging

from visidata import VisiData, Sheet, Progress, ColumnItem, anytype, vd

SASTypes = {
    'string': str,
    'number': float,
}

@VisiData.api
def open_xpt(vd, p):
    return XptSheet(p.base_stem, source=p)

@VisiData.api
def open_sas7bdat(vd, p):
    return SasSheet(p.base_stem, source=p)

class XptSheet(Sheet):
    def iterload(self):
        xport = vd.importExternal('xport')
        xport.v56 = vd.importExternal('xport.v56', 'xport>=3')
        with open(self.source, 'rb') as fp:
            self.library = xport.v56.load(fp)

            self.columns = []
            dataset = self.library[list(self.library.keys())[0]]

            varnames = dataset.contents.Variable.values
            types = dataset.contents.Type.values

            for i, (varname, typestr) in enumerate(zip(varnames, types)):
                self.addColumn(ColumnItem(varname, i, type=float if typestr == 'Numeric' else str))

            for row in dataset.values:
                yield list(row)


class SasSheet(Sheet):
    def iterload(self):
        sas7bdat = vd.importExternal('sas7bdat')
        self.dat = sas7bdat.SAS7BDAT(str(self.source), skip_header=True, log_level=logging.CRITICAL)
        self.columns = []
        for col in self.dat.columns:
            self.addColumn(ColumnItem(col.name.decode('utf-8'), col.col_id, type=SASTypes.get(col.type, anytype)))

        with self.dat as fp:
            yield from Progress(fp, total=self.dat.properties.row_count)




import pytest

from visidata import Sheet, date


class TestVisidataDate:
    def test_date(self):  #1507
        dt = Sheet().customdate('%d%m%Y')
        assert not date(2021, 7, 1) <= dt('22092017')
        assert date(2021, 7, 1) <= dt('28092021')




from visidata import vd, TableSheet
import pytest


class TestMenu:
    def test_menuitems(self):
        vd.addMenuItems('''Column > Add column > foobar > addcol-sparkline''')

        m = TableSheet().getMenuItem(['Column', 'Add column', 'foobar'])
        assert m

        with pytest.raises(AssertionError):
            vd.addMenuItems('''Column > Add column > non-command''')



from visidata import Sheet, VisiData, TypedWrapper, anytype, date, vlen, Column, vd
from collections import defaultdict


@VisiData.api
def open_parquet(vd, p):
    return ParquetSheet(p.base_stem, source=p)


class ParquetColumn(Column):
    def calcValue(self, row):
        val = self.source[row["__rownum__"]]
        if val.type == 'large_string':
            return memoryview(val.as_buffer())[:2**20].tobytes().decode('utf-8')
        else:
            return val.as_py()


class ParquetSheet(Sheet):
    # rowdef: {'__rownum__':int, parquet_col:overridden_value, ...}
    def iterload(self):
        pa = vd.importExternal("pyarrow", "pyarrow")
        pq = vd.importExternal("pyarrow.parquet", "pyarrow")
        from visidata.loaders.arrow import arrow_to_vdtype

        if self.source.is_dir():
            self.tbl = pq.read_table(str(self.source))
        else: 
            with self.source.open('rb') as f:
                self.tbl = pq.read_table(f)

        self.columns = []
        for colname, col in zip(self.tbl.column_names, self.tbl.columns):
            c = ParquetColumn(colname,
                              type=arrow_to_vdtype(col.type),
                              source=col,
                              cache=(col.type.id == pa.lib.Type_LARGE_STRING))
            self.addColumn(c)

        for i in range(self.tbl.num_rows):
            yield dict(__rownum__=i)


@VisiData.api
def save_parquet(vd, p, sheet):
    pa = vd.importExternal("pyarrow")
    pq = vd.importExternal("pyarrow.parquet", "pyarrow")

    typemap = {
        anytype: pa.string(),
        int: pa.int64(),
        vlen: pa.int64(),
        float: pa.float64(),
        str: pa.string(),
        date: pa.date64(),
        # list: pa.array(),
    }

    for t in vd.numericTypes:
        if t not in typemap:
            typemap[t] = pa.float64()

    databycol = defaultdict(list)  # col -> [values]

    for typedvals in sheet.iterdispvals(format=False):
        for col, val in typedvals.items():
            if isinstance(val, TypedWrapper):
                val = None

            databycol[col].append(val)

    data = [
        pa.array(vals, type=typemap.get(col.type, pa.string()))
        for col, vals in databycol.items()
    ]

    schema = pa.schema(
        [(c.name, typemap.get(c.type, pa.string())) for c in sheet.visibleCols]
    )
    with p.open_bytes(mode="w") as outf:
        with pq.ParquetWriter(outf, schema) as writer:
            writer.write_batch(
                pa.record_batch(data, names=[c.name for c in sheet.visibleCols])
            )



'''
Load new table from system clipboard
'''

from visidata import vd, BaseSheet, Path


@BaseSheet.api
def open_syspaste(sheet, filetype='tsv'):
    import io

    v = vd.sysclipValue().strip() or vd.fail('nothing to open')

    p = Path('syspaste'+'.'+filetype, fp=io.BytesIO(v.encode('utf-8')))
    return vd.openSource(p, filetype=filetype)


BaseSheet.addCommand('gShift+P', 'open-syspaste', 'vd.push(open_syspaste(filetype=vd.input("paste as filetype: ", value="tsv")))', 'open clipboard as filetype')



import curses

from visidata import vd, VisiData, BaseSheet, Sheet, AttrDict


# registry of mouse events.  cleared before every draw cycle.
vd.mousereg = []  # list of AttrDict(y=, x=, h=, w=, buttonfuncs=dict)

# sheet mouse position for current mouse event
BaseSheet.init('mouseX', int)
BaseSheet.init('mouseY', int)


@VisiData.after
def initCurses(vd):
    if not getattr(curses, 'mousemask', None):
      return
    curses.MOUSE_ALL = 0xffffffff
    curses.mousemask(curses.MOUSE_ALL if vd.options.mouse_interval else 0)
    curses.def_prog_mode()
    curses.mouseinterval(vd.options.mouse_interval)
    curses.mouseEvents = {}

    for k in dir(curses):
        if k.startswith('BUTTON') or k in ('REPORT_MOUSE_POSITION', '2097152'):
            curses.mouseEvents[getattr(curses, k)] = k


@VisiData.after
def clearCaches(vd):
    vd.mousereg = []


@VisiData.api
def onMouse(vd, scr, x, y, w, h, **kwargs):
    px, py = vd.getrootxy(scr)
    e = AttrDict(x=x+px, y=y+py, w=w, h=h, buttonfuncs=kwargs)
    vd.mousereg.append(e)


@VisiData.api
def getMouse(vd, _x, _y, button):
    for reg in vd.mousereg[::-1]:
        if reg.x <= _x < reg.x+reg.w and reg.y <= _y < reg.y+reg.h and button in reg.buttonfuncs:
            return reg.buttonfuncs[button]


@VisiData.api
def parseMouse(vd, **kwargs):
    'Return list of mouse interactions (clicktype, y, x, name, scr) for curses screens given in kwargs as name:scr.'

    devid, x, y, z, bstate = curses.getmouse()

    clicktype = ''
    if bstate & curses.BUTTON_CTRL:
        clicktype += "Ctrl+"
        bstate &= ~curses.BUTTON_CTRL
    if bstate & curses.BUTTON_ALT:
        clicktype += "Alt+"
        bstate &= ~curses.BUTTON_ALT
    if bstate & curses.BUTTON_SHIFT:
        clicktype += "Shift+"
        bstate &= ~curses.BUTTON_SHIFT

    keystroke = clicktype + curses.mouseEvents.get(bstate, str(bstate))
    ret = AttrDict(keystroke=keystroke, y=y, x=x, found=[])
    for winname, winscr in kwargs.items():
        if not winscr:
            continue
        px, py = vd.getrootxy(winscr)
        mh, mw = winscr.getmaxyx()
        if py <= y < py+mh and px <= x < px+mw:
            ret.found.append(winname)
#            vd.debug(f'{keystroke} at ({x-px}, {y-py}) in window {winname} {winscr}')

    return ret


@VisiData.api
def handleMouse(vd, sheet):
    try:
        vd.keystrokes = ''
        pct = vd.windowConfig['pct']
        topPaneActive = ((vd.activePane == 2 and pct < 0)  or (vd.activePane == 1 and pct > 0))
        bottomPaneActive = ((vd.activePane == 1 and pct < 0)  or (vd.activePane == 2 and pct > 0))
        r = None
        r = vd.parseMouse(top=vd.winTop, bot=vd.winBottom, menu=vd.scrMenu)
        if (bottomPaneActive and 'top' in r.found) or (topPaneActive and 'bot' in r.found):
            vd.activePane = 1 if vd.activePane == 2 else 2
            sheet = vd.activeSheet

        f = vd.getMouse(r.x, r.y, r.keystroke)
        winx, winy = vd.getrootxy(sheet._scr)
        sheet.mouseX, sheet.mouseY = r.x-winx, r.y-winy
        if f:
            if isinstance(f, str):
                if f.startswith('onclick'):
                    if '://' in f:
                        vd.launchBrowser(f[8:])
                    else:
                        sheet.execCommand(f[8:])
                else:
                    for cmd in f.split():
                        sheet.execCommand(cmd)
            else:
                f(r.y, r.x, r.keystroke)

            vd.keystrokes = vd.prettykeys(r.keystroke)
            return ''  #  handled
    except curses.error:
        pass

    return r.keystroke if r else ''


@Sheet.api
def visibleColAtX(sheet, x):
    for vcolidx, (colx, w) in sheet._visibleColLayout.items():
        if colx <= x <= colx+w:
            return vcolidx


@Sheet.api
def visibleRowAtY(sheet, y):
    for rowidx, (rowy, h) in sheet._rowLayout.items():
        if rowy <= y <= rowy+h-1:
            return rowidx


@Sheet.command('BUTTON1_PRESSED', 'go-mouse', 'set cursor to row and column where mouse was clicked')
def go_mouse(sheet):
    ridx = sheet.visibleRowAtY(sheet.mouseY)
    if ridx is not None:
        sheet.cursorRowIndex = ridx
    cidx = sheet.visibleColAtX(sheet.mouseX)
    if cidx is not None:
        sheet.cursorVisibleColIndex = cidx

Sheet.addCommand(None, 'scroll-mouse', 'sheet.topRowIndex=cursorRowIndex-mouseY+1', 'scroll to mouse cursor location')

Sheet.addCommand('ScrollUp', 'scroll-up', 'cursorDown(options.scroll_incr); sheet.topRowIndex += options.scroll_incr', 'scroll one row up')
Sheet.addCommand('ScrollDown', 'scroll-down', 'cursorDown(-options.scroll_incr); sheet.topRowIndex -= options.scroll_incr', 'scroll one row down')



from visidata import vd, VisiData, BaseSheet, Sheet, Column, AttrColumn, ItemColumn, setitem, asyncthread


class IndexSheet(Sheet):
    'Base class for tabular sheets with rows that are Sheets.'
    guide = '''
    # Index Sheet
    This is a list of sheets from `{sheet.source}`.

    - `Enter` to open {sheet.cursorRow}
    - `g Enter` to open all selected sheets
    '''
    rowtype = 'sheets'  # rowdef: Sheet

    columns = [
        Column('name', getter=lambda c,r: r.names[-1], setter=lambda c,r,v: setitem(r.names, -1, v)),
        AttrColumn('rows', 'nRows', type=int, width=9),
        AttrColumn('cols', 'nCols', type=int),
        AttrColumn('keys', 'keyColNames'),
        AttrColumn('source'),
    ]
    nKeys = 1

    def newRow(self):
        return Sheet('', columns=[ItemColumn('', 0)], rows=[])

    def openRow(self, row):
        return row  # rowdef is Sheet

    def getSheet(self, k):
        for vs in self.rows:
            if vs.name == k:
                return vs

    def addRow(self, sheet, **kwargs):
        super().addRow(sheet, **kwargs)
        if not self.options.load_lazy and not sheet.options.load_lazy:
            sheet.ensureLoaded()

    @asyncthread
    def reloadSheets(self, sheets):
        for vs in vd.Progress(sheets):
            vs.reload()


class SheetsSheet(IndexSheet):
    columns = [
        AttrColumn('name'),
        AttrColumn('type', '__class__.__name__'),
        AttrColumn('pane', type=int),
        Column('shortcut', getter=lambda c,r: getattr(r, 'shortcut'), setter=lambda c,r,v: setattr(r, '_shortcut', v)),
        AttrColumn('nRows', type=int),
        AttrColumn('nCols', type=int),
        AttrColumn('nVisibleCols', type=int),
        AttrColumn('cursorDisplay'),
        AttrColumn('keyColNames'),
        AttrColumn('source'),
        AttrColumn('progressPct'),
#        AttrColumn('threads', 'currentThreads', type=vlen),
    ]
    precious = False
    nKeys = 1
    def reload(self):
        self.rows = self.source

    def sort(self):
        self.rows[1:] = sorted(self.rows[1:], key=self.sortkey)


class GlobalSheetsSheet(SheetsSheet):  #1620
    def sort(self):
        IndexSheet.sort(self)


@VisiData.lazy_property
def sheetsSheet(vd):
    return SheetsSheet("sheets", source=vd.sheets)


@VisiData.lazy_property
def allSheetsSheet(vd):
    return GlobalSheetsSheet("sheets_all", source=vd.allSheets)



@Sheet.api
def nextRow(sheet, n=1):
    sheet.cursorRowIndex += n
    sheet.checkCursor()
    return sheet.rows[sheet.cursorRowIndex]  # cursorRow itself might be cached


vd.addCommand('S', 'sheets-stack', 'vd.push(vd.sheetsSheet)', 'open Sheets Stack: join or jump between the active sheets on the current stack')
vd.addCommand('gS', 'sheets-all', 'vd.push(vd.allSheetsSheet)', 'open Sheets Sheet: join or jump between all sheets from current session')

BaseSheet.addCommand('g>', 'open-source-next', 'vd.replace(openSource(source.nextRow())) if isinstance(source, IndexSheet) else fail("parent sheet must be Index Sheet")', 'open next sheet on parent index sheet')
BaseSheet.addCommand('g<', 'open-source-prev', 'vd.replace(openSource(source.nextRow(-1))) if isinstance(source, IndexSheet) else fail("parent sheet must be Index Sheet")', 'open prev sheet on parent index sheet')

IndexSheet.addCommand('g^R', 'reload-selected', 'reloadSheets(selectedRows or rows)', 'reload all selected sheets')

# when diving into a sheet, remove the index unless it is precious
IndexSheet.addCommand('gC', 'columns-selected', 'vd.push(ColumnsSheet("all_columns", source=selectedRows))', 'open Columns Sheet with all visible columns from selected sheets')
IndexSheet.addCommand('^C', 'cancel-row', 'cancelThread(*cursorRow.currentThreads)', 'abort async thread for current sheet')
IndexSheet.addCommand('gz^C', 'cancel-rows', 'for vs in selectedRows: cancelThread(*vs.currentThreads)', 'abort async threads for selected sheets')
SheetsSheet.addCommand('Enter', 'open-row', 'dest=cursorRow; vd.sheets.remove(sheet) if not sheet.precious else None; vd.push(openRow(dest))', 'open sheet referenced in current row')

vd.addGlobals(IndexSheet=IndexSheet,
              SheetsSheet=SheetsSheet,
              GlobalSheetsSheet=GlobalSheetsSheet)



'ASCII monochrome theme; default colors only'

from visidata import vd


vd.themes['asciimono'] = dict(
    disp_note_none='',
    disp_truncator='>',
    disp_oddspace='.',
    disp_more_left='<',
    disp_more_right='>',
    disp_error_val='',
    disp_ambig_width=1,

    disp_pending='',
    disp_note_pending=':',
    disp_note_fmtexc='?',
    disp_note_getexc='!',
    disp_note_typeexc='!',

    color_note_pending='bold',
    color_note_type='',
    color_note_row='',

    disp_column_sep='|',
    disp_keycol_sep='|',
    disp_rowtop_sep='|',
    disp_rowmid_sep='|',
    disp_rowbot_sep='|',
    disp_rowend_sep='|',
    disp_keytop_sep='|',
    disp_keymid_sep='|',
    disp_keybot_sep='|',
    disp_endtop_sep='|',
    disp_endmid_sep='|',
    disp_endbot_sep='|',
    disp_selected_note='+',
    disp_sort_asc='^^^^^^',
    disp_sort_desc='vvvvvv',
    color_default='',
    color_default_hdr='bold',
    color_bottom_hdr='underline',
    color_current_row='reverse',
    color_current_col='bold',
    color_current_hdr='bold reverse',
    color_column_sep='',
    color_key_col='bold',
    color_hidden_col='8',
    color_selected_row='',
    color_edit_cell='',
    color_edit_unfocused='',
    color_graph_hidden='',
    color_graph_selected='bold',
    color_status_replay='',
    color_currency_neg='',
    color_match='',
    color_cmdpalette='',

    color_graph_axis='bold',
    color_sidebar='reverse',
    color_change_pending='reverse',
    color_delete_pending='underline',

    disp_rstatus_fmt=' {sheet.longname} {sheet.nRows:9d} {sheet.rowtype} {sheet.modifiedStatus} {sheet.options.disp_selected_note}{sheet.nSelectedRows}',
    disp_status_fmt='{sheet.shortcut}> {sheet.name}| ',
    disp_lstatus_max=0,
    disp_status_sep=' | ',
    color_keystrokes='bold reverse',
    color_status='bold reverse',
    color_error='bold',
    color_warning='bold',
    color_top_status='underline',
    color_active_status='reverse',
    color_inactive_status='8',
    color_working='',
    color_longname='',
    color_highlight_status='',
    color_sidebar_title='',
    color_heading='',
    color_guide_unwritten='',
    color_code='',

    color_menu='reverse',
    color_menu_active='',
    color_menu_spec='reverse',
    color_menu_help='reverse',
    color_add_pending='',
    disp_menu_boxchars='||--    ||',
    disp_menu_more='>',
    disp_menu_push='+',
    disp_menu_input='_',
    disp_menu_fmt='_.;"`\\ ASCII mono /\':._',
    plot_colors = 'white',
)



"""
# Usage

This plugin normalizes column names in any given sheet, so that the names are:

- Composed only of lowercase letters, numbers, and underscores.

- Valid Python identifiers. This is mostly handled by the rule above, but also
  prohibits names beginning with a digit; that is handled by prefixing those
  names with an underscore.

- Unique within the sheet. Non-unique names are suffixed with "__" and an
  integer.

Unnamed columns are left as such.

For instance, a sheet with the following columns names:

- "Genus, Species"
- "Height"
- "5-score"
- "Height"
- ""
- ""

... would be converted to have the following column names:

- "genus_species"
- "height__0"
- "_5_score"
- "height__1"
- ""
- ""

## Commands

- `normalize-col-names` normalizes the names of all *non-hidden* columns in the
  active sheet, per the approach described above.

"""

__author__ = "Jeremy Singer-Vine <jsvine@gmail.com>"

from visidata import vd, Sheet, asyncthread, Progress
from collections import Counter
import re
import string

nonalphanum_pat = re.compile(r"[^a-z0-9]+")


def normalize_name(name):
    """
    Given a string, return a normalized string, per the first two rules
    described above.
    """
    # Lowercase and replace all non-alphanumeric characters with _
    subbed = re.sub(nonalphanum_pat, "_", name.lower())

    # Remove leading and trailing _s
    stripped = subbed.strip("_")

    # To ensure it's a valid Python identifier
    if (stripped or "_")[0] in string.digits:
        stripped = "_" + stripped

    return stripped


def gen_normalize_names(names):
    """
    Given a list of strings, yield fully-normalized conversions of those
    strings, ensuring that each is unique.
    """
    base = list(map(normalize_name, names))
    counts = Counter(base)

    # Append __{i} to non-unique names
    seen = dict((key, 0) for key in counts.keys())
    for name in base:
        if counts[name] == 1 or name == "":
            norm_name = name
        else:
            norm_name = name + "__" + str(seen[name])
            seen[name] += 1
        yield norm_name


@Sheet.api
@asyncthread
def normalize_column_names(sheet):
    """
    Normalize the names of all non-hidden columns on the active sheet.
    """

    init_names = {}
    gen = gen_normalize_names(c.name for c in sheet.visibleCols)
    prog = Progress(gen, gerund="normalizing", total=sheet.nVisibleCols)

    for i, norm_name in enumerate(prog):
        col = sheet.visibleCols[i]
        init_names[col] = col.name  # Store for undo
        col.name = norm_name

    @asyncthread
    def undo():
        for c, oldname in init_names.items():
            c.name = oldname

    vd.addUndo(undo)


# Add longname-commands to VisiData to execute these methods
Sheet.addCommand(None, "normalize-col-names", "vd.sheet.normalize_column_names()", "normalize the names of all non-hidden columns")

vd.addMenuItems('''
    Column > Rename > normalize all > normalize-col-names
''')



from visidata import vd, Sheet, Column

vd.option('disp_currency_fmt', '%.02f', 'default fmtstr to format for currency values', replay=True, help=vd.help_float_fmt)
vd.theme_option('color_currency_neg', 'red', 'color for negative values in currency displayer', replay=True)


floatchars='+-0123456789.'

@vd.numericType('$')
def currency(*args):
    'dirty float (strip non-numeric characters)'
    if args and isinstance(args[0], str):
        args = [''.join(ch for ch in args[0] if ch in floatchars)]
    return float(*args)


@Column.api
def displayer_currency(col, dw, width=None):
    text = dw.text

    if isinstance(dw.typedval, (int, float)):
        if dw.typedval < 0:
            text = f'({dw.text[1:]})'.rjust(width-1)
            yield ('currency_neg', '')
        else:
            text = text.rjust(width-2)

    yield ('', text)

Sheet.addCommand('$', 'type-currency', 'cursorCol.type=currency; cursorCol.displayer="currency"', 'set type of current column to currency')

vd.addMenuItems('''
    Column > Type as > dirty float > type-currency
''')



#
# Usage: $0 [<options>] [<input> ...]
#        $0 [<options>] --play <cmdlog> [--batch] [-w <waitsecs>] [-o <output>] [field=value ...]

__version__ = '3.3'
__version_info__ = 'saul.pw/VisiData v' + __version__

from copy import copy
import os
import io
import sys
import locale
import datetime
import functools
import signal
import warnings
import builtins  # to override print

from visidata import vd, options, run, BaseSheet, AttrDict, stacktrace
from visidata import Path
from visidata.settings import _get_config_file
import visidata

vd.version_info = __version_info__

vd.option('config', _get_config_file(), 'config file to exec in Python', sheettype=None)
vd.option('play', '', 'file.vdj to replay')
vd.option('batch', False, 'replay in batch mode (with no interface and all status sent to stdout)')
vd.option('output', None, 'save the final visible sheet to output at the end of replay')
vd.option('preplay', '', 'longnames to preplay before replay')
vd.option('imports', 'plugins', 'imports to preload before .visidatarc (command-line only)')
vd.option('nothing', False, 'no config, no plugins, nothing extra')
vd.option('interactive', False, 'run interactive mode after batch replay')

# for --play
def eval_vd(logpath, *args, **kwargs):
    'Instantiate logpath with args/kwargs replaced and replay all commands.'
    log = logpath.read_text()
    if args or kwargs:
        if logpath.ext in ['vdj', 'json', 'jsonl'] or logpath is vd.stdinSource:
            from string import Template
            log = Template(log).safe_substitute(**kwargs)
        else:
            log = log.format(*args, **kwargs)

    src = Path(logpath.given, fptext=io.StringIO(log), filesize=len(log))
    if logpath is vd.stdinSource:
        # replay from stdin only supports .vdj
        vs = vd.openSource(src, filetype='vdj')
    else:
        vs = vd.openSource(src, filetype=src.ext)
    vs.name += '_vd'
    vd.sync(vs.reload())
    vs.vd = vd
    return vs


def duptty():
    'Duplicate stdin/stdout for input/output and reopen tty as stdin/stdout.  Return (stdin, stdout).'
    try:
        fin = open('/dev/tty')
        fout = open('/dev/tty', mode='w')
        stdin = open(os.dup(0),
                     encoding=vd.options.getonly('encoding', 'global', 'utf-8'),
                     errors=vd.options.getonly('encoding_errors', 'global', 'surrogateescape'))  #2047
        stdout = open(os.dup(1))  # for dumping to stdout from interface
        os.dup2(fin.fileno(), 0)
        os.dup2(fout.fileno(), 1)

        # close file descriptors for original stdin/stdout
        fin.close()
        fout.close()
    except Exception as e:
        stdin = sys.stdin
        stdout = sys.stdout

    return stdin, stdout

vd.optalias('i', 'interactive')
vd.optalias('N', 'nothing')
vd.optalias('f', 'filetype')
vd.optalias('p', 'play')
vd.optalias('b', 'batch')
vd.optalias('P', 'preplay')
vd.optalias('o', 'output')
vd.optalias('w', 'replay_wait')
vd.optalias('d', 'delimiter')
vd.optalias('c', 'config')
vd.optalias('r', 'dir_depth', 100000)


@visidata.VisiData.api
def parsePos(vd, arg:str, inputs=None):
    'Return (startsheets:list, startrow:str, startcol:str) from *arg* like "+sheet:subsheet:col:row".  Empty sheetstr in startsheets means the starting pos applies to all sheets.'
    startsheets, startrow, startcol = [], None, None

    if ':' not in arg:
        return (None, arg, None)

    pos = arg.split(':')
    if len(pos) == 1:
        startsheet = [Path(inputs[-1]).base_stem] if inputs else None
        start_pos = (startsheet, pos[0], None)
    elif len(pos) == 2:
        startsheet = [Path(inputs[-1]).base_stem] if inputs else None
        startrow, startcol = pos
        start_pos = (None, startrow, startcol)
    else:  # if len(pos) >= 3:
        startsheets = pos[:-2]
        startrow, startcol = pos[-2:]
        start_pos = (startsheets, startrow, startcol)

    # index subsheets need to be loaded *after* the cursor indexing
    vd.options.set('load_lazy', True, obj=start_pos[0])

    return start_pos


@visidata.VisiData.api
def outputProgressEvery(vd, sheet, seconds:float=0.5):
    import time
    t0 = time.time()

    while not vd.currentReplay:
        time.sleep(.1)

    while vd.currentReplay:
        t = time.time()
        print(f'\r[{t-t0:.1f}s] ', end='', file=sys.stderr)
        if sheet:
            print(f'{sheet.progressPct}  ', end='', file=sys.stderr)
        sys.stderr.flush()
        time.sleep(seconds)

@visidata.VisiData.api
def moveToPos(vd, sources, startsheets, startrow, startcol):
    sheets = []  # sheets to apply startrow:startcol to
    if not startsheets:
        sheets = sources  # apply row/col to all sheets
    else:
        startsheet = startsheets[0] or sources[-1]
        vs = vd.getSheet(startsheet)
        if not vs:
            vd.warning(f'no sheet "{startsheet}"')
            return

        vd.sync(vs.ensureLoaded())
        vd.clearCaches()
        for startsheet in startsheets[1:]:
            rowidx = vs.getRowIndexFromStr(vd.options.rowkey_prefix + startsheet)
            if rowidx is None:
                vd.warning(f'{vs.name} has no subsheet "{startsheet}"')
                vs = None
                break
            vs = vs.rows[rowidx]
            vd.sync(vs.ensureLoaded())
            vd.clearCaches()
        if vs:
            vd.push(vs)
            sheets = [vs]

    if startrow:
        for vs in sheets:
            if vs:
                vs.moveToRow(startrow) or vd.warning(f'{vs} has no row "{startrow}"')

    if startcol:
        for vs in sheets:
            if vs:
                if not vs.moveToCol(startcol):
                    if startcol.isdigit():
                        vs.moveToCol(int(startcol)) # handle indexing by column number
                    else:
                        vd.warning(f'{vs} has no column "{startcol}"')

def main_vd():
    'Open the given sources using the VisiData interface.'
    if '-v' in sys.argv or '--version' in sys.argv:
        print(vd.version_info)
        return 0
    if '-h' in sys.argv or '--help' in sys.argv:
        print((Path(vd.pkg_resources_files(visidata)) / 'man' / 'vd.txt').open().read())
        return 0
    vd.status(__version_info__)

    try:
        locale.setlocale(locale.LC_ALL, '')
    except locale.Error as e:
        vd.warning(e)

    warnings.showwarning = vd.warning
    vd.printerr = lambda *args: builtins.print(*args, file=sys.stderr)

    flPipedInput = not sys.stdin.isatty()
    flPipedOutput = not sys.stdout.isatty()

    try:
        # workaround for bug in curses.wrapper #899
        # https://stackoverflow.com/questions/31440392/curses-wrapper-messing-up-terminal-after-background-foreground-sequence
        vd.tstp_signal = signal.getsignal(signal.SIGTSTP)
    except Exception:
        vd.tstp_signal = None

    vd.stdinSource = Path('-', fp=None)  # fp filled in below after options parsed for encoding

    # parse args, including +sheetname:subsheet:4:3 starting at row:col on sheetname:subsheet[:...]
    after_config = []
    fmtargs = []
    fmtkwargs = {}
    inputs = []

    i=1
    current_args = {}
    global_args = {}
    flGlobal = True
    optsdone = False

    while i < len(sys.argv):
        arg = sys.argv[i]
        if optsdone:
            # copied from final else: clause below
            inputs.append((arg, copy(current_args)))
            fmtargs.append(arg)
        elif arg in ['--']:
            optsdone = True
        elif arg == '-':
            if not flPipedInput:
                vd.fail('to use stdin as a data source, data must be piped into it')
            inputs.append((vd.stdinSource, copy(current_args)))
        elif arg in ['-g', '--global']:
            flGlobal = True
        elif arg in ['-n', '--nonglobal']:
            flGlobal = False
        elif arg.startswith('-'):
            optname = arg.lstrip('-')
            optval = None
            try:
                optname, optval = optname.split('=', maxsplit=1)
                # convert to type
            except Exception:
                pass

            optname = optname.replace('-', '_')
            optname, optval = vd._resolve_optalias(optname, optval)

            if optval is None:  # missing argument, maybe bool?
                opt = vd.options._get(optname)
                if opt:
                    if type(opt.value) is bool:
                        optval = True
                    else:
                        if i >= len(sys.argv)-1:
                            vd.error(f'"-{optname}" missing argument')

                        optval = sys.argv[i+1]
                        i += 1

            # batch and interactive are only meaningful when applied globally,
            # so exclude them from sheet-specific options. Those would
            # override any later change to vd.options.batch in global settings.
            if optname not in ('batch', 'interactive'):
                current_args[optname] = optval
            if flGlobal:
                global_args[optname] = optval
        elif arg.startswith('+'):  # position cursor at start
            after_config.append((vd.moveToPos, *vd.parsePos(arg[1:], inputs=inputs)))
        elif current_args.get('play', None) and '=' in arg:
            # parse 'key=value' pairs for formatting cmdlog template in replay mode
            k, v = arg.split('=', maxsplit=1)
            fmtkwargs[k] = v
        else:
            inputs.append((arg, copy(current_args)))
            fmtargs.append(arg)

        i += 1

    args = AttrDict(current_args)

    if not args.nothing:
        vd.loadConfigAndPlugins(args)

    for k, v in global_args.items():
        options.set(k, v, obj='global')

    vd._stdin, vd._stdout = duptty()  # always dup stdin/stdout
    vd.stdinSource.fptext = vd._stdin
    vd._stdin.close = vd.nop  #1759

    # fetch motd *after* options parsing/setting
    vd.domotd()

    if options.batch:
        if not vd.options.interactive:
            options.undo = False
            options.quitguard = False
        vd.execAsync = vd.execSync  # disable async

    for cmd in (args.preplay or '').split():
        BaseSheet('').execCommand(cmd)

    if not args.play:
        if flPipedInput and not inputs:  # '|vd' without explicit '-'
            inputs.append((vd.stdinSource, copy(current_args)))

    sources = []
    for p, opts in inputs:
        # filetype is a special option, bc it is needed to construct the specific sheet type
        if ('filetype' in current_args) and ('filetype' not in opts):
            opts['filetype'] = current_args['filetype']

        vs = vd.openSource(p, create=True, **opts) or vd.fail(f'could not open {p}')
        for k, v in current_args.items():  # apply final set of args to sheets specifically on cli, if not set otherwise #573
            if not vs.options.is_set(k, vs):
                vs.options[k] = v

        # log source to cmdlog
        vd.cmdlog.openHook(vs, vs.source)
        sources.append(vs)

    for vs in reversed(sources):
        vd.push(vs, load=False) #1471, 1555

    if not vd.sheets and not args.play and not options.batch:
        if 'filetype' in current_args:
            newfunc = getattr(vd, 'new_' + current_args['filetype'], vd.getGlobals().get('new_' + current_args['filetype']))
            datestr = datetime.date.today().strftime('%Y-%m-%d')
            if newfunc:
                vd.status('creating blank %s' % current_args['filetype'])
                vd.push(newfunc(Path(datestr + '.' + current_args['filetype'])))
            else:
                vd.status('new_%s does not exist, creating new blank sheet' % current_args['filetype'])
                vd.push(vd.newSheet(datestr, 1))
        else:
            vd.push(vd.currentDirSheet)

            # log source to cmdlog
            vd.cmdlog.openHook(vd.currentDirSheet, vd.currentDirSheet.source)

    if not args.play:
        if options.batch:
            if sources:
                vd.push(sources[0])

        for (f, *parms) in after_config:
            f(sources, *parms)

        if not options.batch:
            run(vd.sheets[0])
    else:
        if args.play == '-':
            if vd.stdinSource.fptext.isatty():
                vd.fail('replay commands must come by pipe, not by terminal')
            vdfile = vd.stdinSource
        else:
            vdfile = Path(args.play)

        vs = eval_vd(vdfile, *fmtargs, **fmtkwargs)
        if options.batch:
            if not args.debug:
                vd.outputProgressThread = visidata.VisiData.execAsync(vd, vd.outputProgressEvery, vs, seconds=0.5, sheet=BaseSheet())  #1182
            vd.reloadMacros()
            if vd.replay_sync(vs):  # error
                return 1

            if vd.options.interactive:
                vd.options.batch = False  #2639
                vd.execAsync = lambda *args, vd=vd, **kwargs: visidata.VisiData.execAsync(vd, *args, **kwargs)
                run()
        else:
            vd.replay(vs)
            run()

    if vd.stackedSheets and (flPipedOutput or args.output):
        outpath = Path(args.output or '-')
        vd.saveSheets(outpath, vd.activeSheet, confirm_overwrite=False)

    saver_threads = [t for t in vd.unfinishedThreads if t.name.startswith('save_')]
    if saver_threads:
        vd.printerr('finishing %d savers' % len(saver_threads))
        vd.sync(*saver_threads)

    vd._stdout.flush()

    return 0

def vd_cli():
    rc = -1
    try:
        rc = main_vd()
    except BrokenPipeError:
        os.dup2(os.open(os.devnull, os.O_WRONLY), sys.stdout.fileno()) # handle broken pipe gracefully
    except visidata.ExpectedException as e:
        if vd.options.debug:
            raise
    except FileNotFoundError as e:
        print(e, file=sys.stderr)
        if options.debug:
            raise
    except Exception as e:
        for l in stacktrace(): #show the stack trace without carets
            print(l, file=sys.stderr)

    sys.stderr.flush()
    sys.stdout.flush()
    os._exit(rc)  # cleanup can be expensive with large datasets



# to anonymize a column in vd: do "setcol-fake" with e.g. 'name' 'isbn10' or any of the functions on Faker()

import json

from visidata import vd, Column, Sheet, asyncthread, Progress, VisiData

vd.option('faker_locale', 'en_US', 'default locale to use for Faker', replay=True)
vd.option('faker_extra_providers', None, 'list of additional Provider classes to load via add_provider()', replay=True)
vd.option('faker_salt', '', 'Use a non-empty string to enable deterministic fakes')

def addFakerProviders(fake, providers):
    '''
    Add custom providers to Faker. Provider classes typically derive from
    faker.providers.BaseProvider, so check for that here. This helps to
    highlight likely misconfigurations instead of hiding them.

    See also: https://faker.readthedocs.io/en/master/communityproviders.html

    fake: Faker object
    providers: List of provider classes to add
    '''
    faker = vd.importExternal('faker', 'Faker')
    if isinstance(providers, str):
        providers = [ getattr(faker.providers, p) for p in providers.split() ]

    if not isinstance(providers, list):
        vd.fail('options.faker_extra_providers must be a list')

    for provider in providers:
        if not issubclass(provider, faker.providers.BaseProvider):
            vd.warning('"{}" not a Faker Provider'.format(provider.__name__))
            continue
        fake.add_provider(provider)

@Column.api
@asyncthread
def setValuesFromFaker(col, faketype, rows):
    faker = vd.importExternal('faker', 'Faker')
    fake = faker.Faker(col.sheet.options.faker_locale)
    if col.sheet.options.faker_extra_providers:
        addFakerProviders(fake, col.sheet.options.faker_extra_providers)
    fakefunc = getattr(fake, faketype, None) or vd.fail(f'no such faker "{faketype}"')

    fakeMap = {}
    fakeMap[None] = None
    fakeMap[col.sheet.options.null_value] = col.sheet.options.null_value

    vd.addUndoSetValues([col], rows)
    salt = col.sheet.options.faker_salt

    for r in Progress(rows):
        v = col.getValue(r)
        if v in fakeMap:
            newv = fakeMap[v]
        else:
            if salt:
                # Reset the Faker seed for each value. For a given salt string,
                # the same cell value will always generate the same fake value.
                fake.seed_instance(json.dumps(v) + salt)
            newv = fakefunc()
            fakeMap[v] = newv
        col.setValue(r, newv)


Sheet.addCommand(None, 'setcol-fake', 'cursorCol.setValuesFromFaker(input("faketype: ", type="faketype"), selectedRows)', 'replace values in current column for selected rows with fake values')



import builtins
import contextlib
import os
import curses
import signal
import threading
import time

from visidata import vd, VisiData, colors, ESC, options, BaseSheet, AttrDict

__all__ = ['ReturnValue', 'run']

vd.curses_timeout = 100 # curses timeout in ms
vd.timeouts_before_idle = 10
vd.min_draw_ms = 100  # draw_all at least this often, even if keystrokes are pending
vd._lastDrawTime = 0  # last time drawn (from time.time())


class ReturnValue(BaseException):
    'raise ReturnValue(ret) to exit from an inner runresult() with its result.'
    pass


@VisiData.api
def callNoExceptions(vd, func, *args, **kwargs):
    'Catch and log any raised exceptions.  Reraise when options.debug.'
    try:
        return func(*args, **kwargs)
    except Exception as e:
        vd.exceptionCaught(e)


@VisiData.api
def drawSheet(vd, scr, sheet):
    'Erase *scr* and draw *sheet* on it, including status bars and sidebar.'

    sheet.ensureLoaded()

    if not scr:
        return

    scr.erase()  # clear screen before every re-draw
    scr.bkgd(' ', colors.color_default.attr)

    sheet._scr = scr

    vd.callNoExceptions(sheet.draw, scr)
    vd.callNoExceptions(vd.drawLeftStatus, scr, sheet)
    vd.callNoExceptions(vd.drawRightStatus, scr, sheet)  # visible during this getkeystroke


vd.windowConfig = dict(pct=0, n=0, h=0, w=0)  # n=top line of bottom window; h=height of bottom window; w=width of screen

vd.winTop = None
vd.scrMenu = None
vd.scrFull = None


@VisiData.api
def setWindows(vd, scr, pct=None):
    'Assign winTop, winBottom, win1 and win2 according to options.disp_splitwin_pct.'
    if pct is None:
        pct = options.disp_splitwin_pct  # percent of window for secondary sheet (negative means bottom)
    disp_menu = getattr(vd, 'menuRunning', None) or vd.options.disp_menu
    topmenulines = 1 if disp_menu else 0
    h, w = scr.getmaxyx()
    if h == 1: topmenulines = 0

    n = 0
    if pct:
        # on 100 line screen, pct = 25 means second window on lines 75-100.  pct -25 -> lines 0-25
        n = abs(pct)*h//100
        n = min(n, h-topmenulines-3)
        n = max(3, n)
        if n > h: n = 0

    desiredConfig = dict(pct=pct, n=n, h=h-topmenulines, w=w)

    if vd.scrFull is not scr or vd.windowConfig != desiredConfig:
        if not topmenulines:
            vd.scrMenu = None
        elif not vd.scrMenu:
            vd.scrMenu = vd.subwindow(scr, 0, 0, w, h)
            vd.scrMenu.keypad(1)

        vd.winTop = vd.subwindow(scr, 0, topmenulines, w, n)
        vd.winTop.keypad(1)
        vd.winBottom = vd.subwindow(scr, 0, n+topmenulines, w, h-n-topmenulines)
        vd.winBottom.keypad(1)
        if pct == 0 or pct >= 100:  # no second pane
            vd.win1 = vd.winBottom
            # drawing to 0-line window causes problems
            vd.win2 = None
        elif pct > 0: # pane 2 from line n to bottom
            vd.win1 = vd.winTop
            vd.win2 = vd.winBottom
        elif pct < 0: # pane 2 from line 0 to n
            vd.win1 = vd.winBottom
            vd.win2 = vd.winTop

        for vs in vd.sheetstack(1)[0:1]+vd.sheetstack(2)[0:1]:
            vs.refresh()

        vd.windowConfig = desiredConfig
        vd.scrFull = scr
        return True


@VisiData.api
def draw_all(vd):
    'Draw all sheets in all windows.'
    vd.clearCaches()

    ss1 = vd.sheetstack(1)
    ss2 = vd.sheetstack(2)
    if ss1 and not ss2:
        vd.activePane = 1
        vd.setWindows(vd.scrFull)
        vd.drawSheet(vd.win1, ss1[0])
        if vd.win2:
            vd.win2.erase()
    elif not ss1 and ss2:
        vd.activePane = 2
        vd.setWindows(vd.scrFull)
        vd.drawSheet(vd.win2, ss2[0])
        if vd.win1:
            vd.win1.erase()
    elif ss1 and ss2 and vd.win2:
        vd.drawSheet(vd.win1, ss1[0])
        vd.drawSheet(vd.win2, ss2[0])
    elif ss1 and ss2 and not vd.win2:
        vd.drawSheet(vd.win1, vd.sheetstack(vd.activePane)[0])
        vd.setWindows(vd.scrFull)

    if vd.scrMenu:
        vd.callNoExceptions(vd.drawMenu, vd.scrMenu, vd.activeSheet)

    vd.callNoExceptions(vd.drawSidebar, vd.scrFull, vd.activeSheet)

    if vd.win1:
        vd.win1.refresh()
    if vd.win2:
        vd.win2.refresh()
    if vd.scrMenu:
        vd.scrMenu.refresh()


@VisiData.api
def runresult(vd):
    try:
        err = vd.mainloop(vd.scrFull)
        if err:
            raise Exception(err)
    except ReturnValue as e:
        return e.args[0]


@VisiData.api
def mainloop(vd, scr):
    'Manage execution of keystrokes and subsequent redrawing of screen.'
    nonidle_timeout = vd.curses_timeout

    scr.timeout(vd.curses_timeout)
    with contextlib.suppress(curses.error):
        curses.curs_set(0)

    numTimeouts = 0
    prefixWaiting = False
    vd.scrFull = scr
    if not vd.wantsHelp('help'):
        vd.disp_help = -1

    vd.keystrokes = ''
    vd.drawThread = threading.current_thread()
    while True:
        if not vd.stackedSheets and vd.currentReplay is None:
            return

        sheet = vd.activeSheet

        if not sheet:
            continue  # waiting for replay to push sheet

        threading.current_thread().sheet = sheet

        vd.setWindows(vd.scrFull)

        # a newly created sheet needs to be drawn once to set its _scr
        if vd.activeSheet._scr is None or \
           not vd.drainPendingKeys(scr) or \
           time.time() - vd._lastDrawTime > vd.min_draw_ms/1000:  #1459
            vd.draw_all()
            vd._lastDrawTime = time.time()

        keystroke = vd.getkeystroke(scr, sheet)

        if not keystroke and prefixWaiting and "Alt+" in vd.keystrokes:  # timeout ESC
            vd.keystrokes = ''

        if keystroke:  # wait until next keystroke to clear statuses and previous keystrokes
            numTimeouts = 0
            if not prefixWaiting:
                vd.keystrokes = ''

            vd.statuses.clear()

            if keystroke == 'KEY_MOUSE':
                try:
                    keystroke = vd.handleMouse(sheet)  # if it was handled, don't handle again as a regular keystroke
                except Exception as e:
                    vd.exceptionCaught(e)

            if keystroke and keystroke in vd.allPrefixes and keystroke in vd.keystrokes[:-1]:
                vd.warning('duplicate prefix: ' + keystroke)
                vd.keystrokes = ''
            else:
                keystroke = vd.prettykeys(keystroke)
                vd.keystrokes += keystroke

        vd.drawRightStatus(sheet._scr, sheet)  # visible for commands that wait for input

        if not keystroke:  # timeout instead of keypress
            pass
        elif keystroke == 'Ctrl+Q':
            return vd.lastErrors and '\n'.join(vd.lastErrors[-1])
        elif vd.bindkeys._get(vd.keystrokes) is not None:
            sheet.execCommand(vd.keystrokes, keystrokes=vd.keystrokes)
            prefixWaiting = False
        elif vd.keystrokes in vd.allPrefixes:
            prefixWaiting = True
        else:
            vd.status('no command for "%s"' % (vd.keystrokes))
            prefixWaiting = False

        # play next queued command
        if vd._nextCommands and not vd.unfinishedThreads:
            cmd = vd._nextCommands.pop(0)
            if isinstance(cmd, (dict, list)):  # .vd cmdlog rows are NamedListTemplate
                try:
                    if vd.replayOne(cmd):
                        vd.replay_cancel()
                except Exception as e:
                    vd.exceptionCaught(e)
                    vd.replay_cancel()
            else:
                sheet.execCommand(cmd, keystrokes=vd.keystrokes)

        if not vd._nextCommands:
            if vd.currentReplay:
                vd.currentReplayRow = None
                vd.currentReplay = None

        vd.checkForFinishedThreads()
        vd.callNoExceptions(sheet.checkCursor)

        time.sleep(0)  # yield to other threads which may not have started yet
        if vd._nextCommands:
            if vd.unfinishedThreads:  #2369 #2635
                # while running a bg thread for a command, schedule infrequent redraws
                vd.curses_timeout = nonidle_timeout
            else:
                # otherwise, schedule the next redraw and command (immediately, for default replay_wait)
                vd.curses_timeout = int(vd.options.replay_wait*1000)
        elif vd.unfinishedThreads:
            vd.curses_timeout = nonidle_timeout
        else:
            numTimeouts += 1
            if vd.timeouts_before_idle >= 0 and numTimeouts >= vd.timeouts_before_idle:
                vd.curses_timeout = -1
            else:
                vd.curses_timeout = nonidle_timeout

        scr.timeout(vd.curses_timeout)


@VisiData.api
def initCurses(vd):
    # reduce ESC timeout to 25ms. http://en.chys.info/2009/09/esdelay-ncurses/
    os.putenv('ESCDELAY', '25')
    curses.use_env(True)

    scr = curses.initscr()

    curses.start_color()

    colors.setup()

    curses.noecho()

    curses.raw()    # get control keys instead of signals
    curses.meta(1)  # allow "8-bit chars"

    scr.keypad(1)

    curses.def_prog_mode()

    vd.drainPendingKeys(scr)
    if '\x1b' in vd.pendingKeys:  #1993
        # if start of an ANSI escape sequence, might be mangled, discard remaining keystrokes
        vd.pendingKeys.clear()
        curses.flushinp()

    return scr


def wrapper(f, *args, **kwargs):
    try:
        scr = vd.initCurses()
        return f(scr, *args, **kwargs)
    finally:
        curses.endwin()


@VisiData.global_api
def run(vd, *sheetlist):
    'Main entry point; launches vdtui with the given sheets already pushed (last one is visible)'

    scr = None
    try:
        # Populate VisiData object with sheets from a given list.
        for vs in sheetlist:
            vd.push(vs, load=False)

        scr = vd.initCurses()
        ret = vd.mainloop(scr)
    except curses.error as e:
        if vd.options.debug:
            raise
        vd.fail(str(e))
    finally:
        if scr:
            curses.endwin()

    vd.cancelThread(*[t for t in vd.unfinishedThreads if not t.name.startswith('save_')])

    if ret:
        builtins.print(ret)

    return ret

@VisiData.api
def addCommand(vd, *args, **kwargs):
    return BaseSheet.addCommand(*args, **kwargs)


import sys
vd.addGlobals({k:getattr(sys.modules[__name__], k) for k in __all__})
